import random
import threading

from core_10x.traitable import Traitable, T, RT, RC, RC_TRUE
from ui_10x.utils import ux, UxAsync
from ui_10x.table_view import TableView


class MarketSymbol(Traitable):
    symbol: str     = RT(T.READONLY)
    bid: float      = RT()
    price: float    = RT()
    ask: float      = RT()
    prev_close: float = RT()

    dp: float           = RT(T.HIDDEN,  default = 0.)
    prev_price: float   = RT(T.HIDDEN)
    delta: float        = RT(T.HIDDEN)
    std: float          = RT(T.HIDDEN)
    delta_mean: float   = RT(T.HIDDEN)

    def delta_mean_get(self) -> float:
        return self.prev_close * 0.001

    def dp_set(self, trait, value) -> RC:
        self.prev_price = self.price
        return self.raw_set_value(trait, value)

    def prev_price_get(self) -> float:
        return self.prev_close

    def price_get(self) -> float:
        return self.prev_price + self.dp

    def price_style_sheet(self) -> str:
        d = self.price - self.prev_close
        if d > 0:
            c = 'green'
        elif d < 0:
            c = 'red'
        else:
            c = 'black'
        return T.fg_color(c)

    def delta_get(self) -> float:
        mm = self.price
        return random.gauss(self.delta_mean, self.std)

    def bid_get(self) -> float:
        mm = self.price
        if mm != None:
            mm -= self.delta
        return mm

    def ask_get(self) -> float:
        mm = self.price
        if mm != None:
            mm += self.delta
        return mm

class MarketMonitor:
    s_symbols = [
        MarketSymbol(symbol = 'MSFT',  prev_close = 398.04,     std = 0.9 ),
        MarketSymbol(symbol = 'MS',    prev_close = 172.70,     std = 0.5 ),
        MarketSymbol(symbol = 'JPM',   prev_close = 307.32,     std = 0.7 ),
        MarketSymbol(symbol = 'IBM',   prev_close = 258.29,     std = 0.15 ),
        MarketSymbol(symbol = 'GS',    prev_close = 918.30,     std = 0.2 ),
        MarketSymbol(symbol = 'GOOG',  prev_close = 302.90,     std = 0.3 ),
        MarketSymbol(symbol = 'NVDA',  prev_close = 188.50,     std = 10.0 ),
        MarketSymbol(symbol = 'BAC',   prev_close = 52.74,      std = 0.9 ),
        MarketSymbol(symbol = 'AMZN',  prev_close = 202.93,     std = 0.8 ),
        MarketSymbol(symbol = 'AAPL',  prev_close = 263.47,     std = 0.7 ),
    ]

    def __init__(self):
        UxAsync.init(self.update_mkt_data)
        self.timer = threading.Timer(3, self.process_timer)
        self.next_item = 0

    def widget(self):
        self.table = table = TableView(MarketSymbol)
        self.timer.start()
        return table

    def update_mkt_data(self):
        i = self.next_item
        if i < len(self.s_symbols):
            self.table.extend_data([ self.s_symbols[i] ])
            self.next_item += 1

        symbols = self.table.model().m_data
        for row, symbol in enumerate(symbols):
            d = random.gauss(symbol.delta_mean, symbol.std)
            if random.randint(0, 10) < 5:
                d = -d
            symbol.dp = d
            self.table.render_traitable(row, symbol)

        dt = random.randint(0, 2)
        self.timer = threading.Timer(dt, self.process_timer)
        self.timer.start()

    def process_timer(self):
        UxAsync.call(self.update_mkt_data)

if __name__ == '__main__':
    from core_10x.exec_control import INTERACTIVE
    from ui_10x.utils import UxDialog

    ux.init()

    with INTERACTIVE():
        mm = MarketMonitor()
        d = UxDialog(mm.widget(), title = 'Enjoy watching some stocks :-)')
        d.exec()




from unittest.mock import MagicMock

import pytest
from core_10x.nucleus import Nucleus
from core_10x.testlib.test_store import TestStore as TraitableStore
from core_10x.trait_filter import GT, f
from core_10x.ts_store import TsStore
from core_10x.ts_union import TsUnion, TsUnionCollection, _OrderKey


@pytest.fixture
def union_collection():
    collection1 = MagicMock()
    collection2 = MagicMock()
    union = TsUnionCollection(collection1, collection2)
    return union, collection1, collection2


def test_empty():
    col = TsUnionCollection()
    assert col.collections == ()

    assert not col.id_exists('1')
    assert not col.exists(f(x=GT(1)))
    assert list(col.find()) == []
    assert col.max('x') is None
    assert col.min('x') is None
    assert col.count() == 0


def test_find(union_collection, args=None):
    union, collection1, collection2 = union_collection
    args = (f(x=GT(1)),) if args is None else args
    collection1.find.return_value = [{Nucleus.ID_TAG(): 2}]
    collection2.find.return_value = [{Nucleus.ID_TAG(): 1}]
    results = list(union.find(*args))
    assert results == [{Nucleus.ID_TAG(): 1}, {Nucleus.ID_TAG(): 2}]
    collection1.find.assert_called_once_with((args and args[0]) or None, _order=None, _at_most=0)
    collection2.find.assert_called_once_with((args and args[0]) or None, _order=None, _at_most=0)


def test_no_args(union_collection):
    test_find(union_collection, args=())


def test_find_empty(union_collection):
    union, collection1, collection2 = union_collection
    collection1.find.return_value = []
    collection2.find.return_value = []
    results = list(union.find())
    assert results == []


def test_save_new(union_collection):
    union, collection1, collection2 = union_collection
    serialized_traitable = {Nucleus.ID_TAG(): 1}
    union.save_new(serialized_traitable)
    collection1.save_new.assert_called_once_with(serialized_traitable, overwrite=False)
    collection2.save.assert_not_called()


def test_save(union_collection):
    union, collection1, collection2 = union_collection
    serialized_traitable = {Nucleus.ID_TAG(): 1}
    collection1.exists.return_value = False
    union.save(serialized_traitable)
    collection1.save_new.assert_called_once_with(serialized_traitable)
    collection1.save.assert_not_called()
    collection2.save_new.assert_not_called()
    collection2.save.assert_not_called()

    collection1.reset_mock()
    collection1.exists.return_value = True
    union.save(serialized_traitable)
    collection1.save_new.assert_not_called()
    collection1.save.assert_called_once_with(serialized_traitable)
    collection2.save_new.assert_not_called()
    collection2.save.assert_not_called()


def test_delete(union_collection):
    union, collection1, collection2 = union_collection
    id_value = '1'
    collection1.delete.return_value = True
    collection1.count.return_value = 0
    collection2.count.return_value = 0
    result = union.delete(id_value)
    assert result
    collection1.delete.assert_called_once_with(id_value)
    collection2.delete.assert_not_called()

    collection2.count.return_value = 1
    result = union.delete(id_value)
    assert not result
    collection2.delete.assert_not_called()


def test_create_index(union_collection):
    union, collection1, collection2 = union_collection
    name = 'index_name'
    trait_name = 'trait_name'
    union.create_index(name, trait_name)
    collection1.create_index.assert_called_once_with(name, trait_name)
    collection2.create_index.assert_not_called()


def test_max(union_collection):
    union, collection1, collection2 = union_collection
    trait_name = 'trait_name'
    collection1.max.return_value = {'trait_name': 1}
    collection2.max.return_value = {'trait_name': 2}
    result = union.max(trait_name)
    assert result == {'trait_name': 2}
    collection1.max.assert_called_once_with(trait_name, None)
    collection2.max.assert_called_once_with(trait_name, None)


def test_max_empty(union_collection):
    union, collection1, collection2 = union_collection
    trait_name = 'trait_name'
    collection1.max.return_value = None
    collection2.max.return_value = None
    result = union.max(trait_name)
    assert result is None


def test_min(union_collection):
    union, collection1, collection2 = union_collection
    trait_name = 'trait_name'
    collection1.min.return_value = {'trait_name': 1}
    collection2.min.return_value = {'trait_name': 2}
    result = union.min(trait_name)
    assert result == {'trait_name': 1}
    collection1.min.assert_called_once_with(trait_name, None)
    collection2.min.assert_called_once_with(trait_name, None)


def test_min_empty(union_collection):
    union, collection1, collection2 = union_collection
    trait_name = 'trait_name'
    collection1.min.return_value = None
    collection2.min.return_value = None
    result = union.min(trait_name)
    assert result is None


def test_multiple_sort_keys(union_collection):
    union, collection1, collection2 = union_collection
    data1 = [
        {'_id': '1', 'group': 'A', 'value': 30},
        {'_id': '2', 'group': 'A', 'value': 10},
        {'_id': '3', 'group': 'B', 'value': 10},
    ]
    data2 = [
        {'_id': '5', 'group': 'A', 'value': 40},
        {'_id': '6', 'group': 'B', 'value': 20},
        {'_id': '7', 'group': 'B', 'value': 20},
    ]
    collection1.find.return_value = data1
    collection2.find.return_value = data2

    # Test multiple sort keys
    results = list(union.find(_order={'group': 1, 'value': -1}))
    sorted_results = [
        {'_id': '5', 'group': 'A', 'value': 40},
        {'_id': '1', 'group': 'A', 'value': 30},
        {'_id': '2', 'group': 'A', 'value': 10},
        {'_id': '6', 'group': 'B', 'value': 20},
        {'_id': '7', 'group': 'B', 'value': 20},
        {'_id': '3', 'group': 'B', 'value': 10},
    ]
    assert results == sorted_results


def test_none_handling(union_collection):
    union, collection1, collection2 = union_collection
    data2 = [
        {'_id': '2', 'value': 20},
        {'_id': '4', 'value': 40},
    ]
    collection1.find.return_value = None
    collection2.find.return_value = data2

    # Test handling None results
    results = list(union.find())
    assert len(results) == 2
    assert [r['_id'] for r in results] == ['2', '4']

    # Test all collections returning None
    collection2.find.return_value = None
    results = list(union.find())
    assert len(results) == 0


def test_sorting(union_collection):
    union, collection1, collection2 = union_collection
    data1 = [
        {'_id': '1', 'value': 30},
        {'_id': '3', 'value': 10},
    ]
    data2 = [
        {'_id': '2', 'value': 40},
        {'_id': '4', 'value': 20},
    ]
    collection1.find.return_value = data1
    collection2.find.return_value = data2

    # Test default sorting by _id ascending
    results = list(union.find())
    assert len(results) == 4
    assert [r['_id'] for r in results] == ['1', '2', '3', '4']

    # Test descending sort by value with limit
    results = list(union.find(_order={'value': -1}, _at_most=2))
    assert [r['value'] for r in results] == [40, 30]

    # Test ascending sort by value with limit
    collection1.find.return_value = reversed(data1)
    collection2.find.return_value = reversed(data2)
    results = list(union.find(_order={'value': 1}, _at_most=3))
    assert len(results) == 3
    assert [r['value'] for r in results] == [10, 20, 30]


def test_sorting_dict(union_collection):
    union, collection1, collection2 = union_collection
    data1 = [
        {'_id': '1', 'dict_field': {'a': 1, 'b': 2}},
        {'_id': '3', 'dict_field': {'a': 2, 'b': 1}},
    ]
    data2 = [
        {'_id': '2', 'dict_field': {'a': 1, 'b': 3}},
        {'_id': '4', 'dict_field': {'a': 2, 'b': 2}},
    ]
    # Test ascending sort by dict field
    collection1.find.return_value = data1
    collection2.find.return_value = data2
    results = list(union.find(_order={'dict_field': 1}))
    assert len(results) == 4
    assert [r['_id'] for r in results] == ['1', '2', '3', '4']

    # Test descending sort by dict field
    collection1.find.return_value = reversed(data1)
    collection2.find.return_value = reversed(data2)
    results = list(union.find(_order={'dict_field': -1}))
    assert len(results) == 4
    assert [r['_id'] for r in results] == ['3', '1', '4', '2']


@pytest.fixture
def union_store():
    mock_store1 = MagicMock()
    mock_store2 = MagicMock()
    union_store = TsUnion(mock_store1, mock_store2)
    return union_store, mock_store1, mock_store2


def test_collection_names(union_store):
    union_store, mock_store1, mock_store2 = union_store
    mock_store1.collection_names.return_value = ['collection1']
    mock_store2.collection_names.return_value = ['collection2']
    result = union_store.collection_names()
    assert set(result) == {'collection1', 'collection2'}
    mock_store1.collection_names.assert_called_once()
    mock_store2.collection_names.assert_called_once()


def test_collection(union_store):
    union_store, mock_store1, mock_store2 = union_store
    collection_name = 'collection_name'
    mock_store1.collection.return_value = MagicMock()
    mock_store2.collection.return_value = MagicMock()
    result = union_store.collection(collection_name)
    assert isinstance(result, TsUnionCollection)
    mock_store1.collection.assert_called_once_with(collection_name)
    mock_store2.collection.assert_called_once_with(collection_name)


def test_delete_collection(union_store):
    union_store, mock_store1, mock_store2 = union_store
    collection_name = 'collection_name'
    union_store.delete_collection(collection_name)
    mock_store1.delete_collection.assert_called_once_with(collection_name)
    mock_store2.delete_collection.assert_not_called()


@pytest.fixture
def ts_union():
    assert not TsUnion.s_instances
    store_spec = dict(driver_name='TEST_DB', hostname='localhost', dbname='dbname1', username='')
    union_store = TsUnion.instance(store_spec, store_spec | dict(dbname='dbname2'))
    yield union_store
    TsUnion.s_instances.clear()


def test_new_instance(ts_union):
    assert isinstance(ts_union, TsUnion)
    assert all(isinstance(store, TraitableStore) for store in ts_union.stores)

    assert sum(1 for v in TsStore.s_instances.values() if isinstance(v, TsUnion)) == 1
    assert sum(1 for v in TsStore.s_instances.values() if isinstance(v, TraitableStore)) == 2

    assert list(TsUnion.s_instances.keys()) == [
        (('dbname', 'dbname1'), ('hostname', 'localhost'), ('username', '')),
        (('dbname', 'dbname2'), ('hostname', 'localhost'), ('username', '')),
        (
            (('dbname', 'dbname1'), ('driver_name', 'TEST_DB'), ('hostname', 'localhost'), ('username', '')),
            (('dbname', 'dbname2'), ('driver_name', 'TEST_DB'), ('hostname', 'localhost'), ('username', '')),
        ),
    ]


def test_equal_same_order():
    assert 0 == _OrderKey._dict_cmp({'a': 1, 'b': 2}, {'a': 1, 'b': 2})


def test_equal_different_order():
    assert 1 == _OrderKey._dict_cmp({'b': 1, 'a': 2}, {'a': 2, 'b': 1})
    assert -1 == _OrderKey._dict_cmp({'a': 1, 'b': 2}, {'b': 2, 'a': 1})


def test_less_key_order():
    assert -1 == _OrderKey._dict_cmp({'a': 1, 'b': 2}, {'b': 2, 'a': 1})  # 'a' < 'b'


def test_greater_key_order():
    assert 1 == _OrderKey._dict_cmp({'b': 2, 'a': 1}, {'a': 1, 'b': 2})  # 'b' > 'a'


def test_less_value():
    assert -1 == _OrderKey._dict_cmp({'a': 1, 'b': 2}, {'a': 1, 'b': 3})


def test_greater_value():
    assert 1 == _OrderKey._dict_cmp({'a': 1, 'b': 3}, {'a': 1, 'b': 2})


def test_shorter_less():
    assert -1 == _OrderKey._dict_cmp({'a': 1}, {'a': 1, 'b': 2})


def test_longer_greater():
    assert 1 == _OrderKey._dict_cmp({'a': 1, 'b': 2}, {'a': 1})


def test_nested_less():
    assert -1 == _OrderKey._dict_cmp({'a': {'x': 1}}, {'a': {'y': 2}})  # 'x' < 'y'


def test_nested_greater():
    assert 1 == _OrderKey._dict_cmp({'a': {'y': 2}}, {'a': {'x': 1}})  # 'y' > 'x'


def test_example_from_query():
    d = {'y': 20, 'x': 10}  # order: y, x
    od = {'x': 10, 'y': 25}  # order: x, y
    assert 1 == _OrderKey._dict_cmp(d, od)  # 'y' > 'x'


def test_counterexample_values_ignored_on_key_diff():
    d = {'a': 100}
    od = {'b': 1}
    assert -1 == _OrderKey._dict_cmp(d, od)  # 'a' < 'b', values ignored


def test_counterexample_early_value_diff():
    d = {'a': 1, 'c': 4}
    od = {'a': 2, 'b': 3}
    assert -1 == _OrderKey._dict_cmp(d, od)  # 1 < 2, later ignored



if __name__ == '__main__':
    from infra_10x.mongodb_store import MongoCollectionHelper, MongoStore

    store = MongoStore.instance('localhost', 'test', '', '')

    id_value = 'AAAA'
    rev = 10

    serialized_traitable = dict(_id=id_value, _rev=rev, name='test', age=60)

    coll = store.collection('test')

    data1 = dict(serialized_traitable)
    pipeline1 = []
    filter1 = {}
    coll.filter_and_pipeline('_id', '_rev', id_value, rev, data1, filter1, pipeline1)

    data2 = dict(serialized_traitable)
    pipeline2 = []
    filter2 = {}

    MongoCollectionHelper.prepare_filter_and_pipeline(data2, filter2, pipeline2)

    assert filter1 == filter2
    assert pipeline1 == pipeline2



class XNoneType:
    """
    XNone is the 10x alternative to None representing a special empty object.
    As opposed to None, XNone may be used in many built-in python operations without extra checking, e.g.

    a_dict = XNone
    for key, value in a_dict.items():
        ...

    The above loop will merely exit on the first iteration.

    - all arithmetic operations with XNone will return XNone
    - all relational operations with XNone will return True or False accordingly
    - a call to XNone(*args, **kwargs) will return XNone
    - XNone is an empty iterator
    - XNone is also an empty context manager
    """

    def __getattr__(self, item):                return self
    def __setattr__(self, key, value):          raise AttributeError('May not setattr to XNone')
    def __getitem__(self, item):                return self
    def __setitem__(self, key, value):          raise TypeError('\'XNone\' object does not support item assignment')
    def __call__(self, *args, **kwargs):        return self
    def __hash__(self):                         return id(self)

    def __int__(self):                          return self
    def __bool__(self):                         return False
    def __float__(self):                        return self

    def __repr__(self):                         return 'XNone'
    def __str__(self):                          return ''

    def __and__(self, other):                   return False
    def __or__(self, other):                    return other
    def __xor__(self, other):                   return other
    def __invert__(self):                       return None   #-- the opposite to XNone
    def __lshift__(self, other):                return self
    def __rshift__(self, other):                return self

    def __len__(self):                          return 0
    def __iter__(self):                         return self
    def __next__(self):                         raise StopIteration

    def __enter__(self):                        return self
    def __exit__(self, exc_type, exc_val, exc_tb):  pass

    def __eq__(self, other):                    return other is self
    def __ne__(self, other):                    return other is not self
    def __lt__(self, other):                    return other is not self
    def __le__(self, other):                    return True
    def __gt__(self, other):                    return False
    def __ge__(self, other):                    return other is self
    def __abs__(self):                          return self
    def __neg__(self):                          return self
    def __add__(self, other):                   return self
    def __radd__(self, other):                  return self
    def __sub__(self, other):                   return self
    def __rsub__(self, other):                  return self
    def __mul__(self, other):                   return self
    def __rmul__(self, other):                  return self
    def __pow__(self, power, modulo = None):    return self
    def __truediv__(self, other):               return self
    def __rtruediv__(self, other):              return self
    def __floordiv__(self, other):              return self
    def __rfloordiv__(self, other):             return self

    s_serialized = chr(8)

    def __init_subclass__(cls, **kwargs):
        raise TypeError('May not derive from XNone')

XNone = XNoneType()



from datetime import date

from core_10x.named_constant import NamedConstant
from core_10x.traitable import THIS_CLASS, AnonymousTraitable, T, Traitable


class MARRITAL_STATUS(NamedConstant):
    SINGLE      = ()
    MARRIED     = ()
    DIVORCED    = ()
    WIDOWED     = ()

class Address(AnonymousTraitable):
    street: str     = T()
    cszip: str      = T()

class Person(Traitable):
    ssn: str                            = T(T.ID)

    dob: date                           = T()
    name: str                           = T()
    gender: bool                        = T()
    marrital_status: MARRITAL_STATUS    = T()

    spouse: THIS_CLASS                  = T()
    children: list                      = T()

    address: Address                    = T(T.EMBEDDED)

class Dict(Traitable):
    payload: dict                       = T()


if __name__ == '__main__':
    from core_10x.manual_tests.traitable_serialization_test import MARRITAL_STATUS, Address, Person

    address = Address(street="145 Austin Dr", cszip="Burlington, VT 05401")
            
    woman       = Person(ssn = '008-59-6666', name = 'Alice Smith',     dob = date(1972, 8, 21), _replace = True)
    daughter    = Person(ssn = '008-77-7777', name = 'Ann Smith',       dob = date(1997, 6, 17), _replace = True)
    son         = Person(ssn = '008-99-5555', name = 'Nathan Smith',    dob = date(1999, 9, 23), _replace = True)


    man     = Person(_replace = True,
        ssn     = '009-87-4444',
        name    = 'John Smith',
        gender  = True,
        dob     = date(1965, 2, 3),
        marrital_status = MARRITAL_STATUS.MARRIED,
        spouse = woman,
        children = [daughter, son],
        address = address
    )

    ss = man.serialize_object()

    from core_10x.package_refactoring import PackageRefactoring

    class_id = PackageRefactoring.find_class_id(Person)
    p_cls = PackageRefactoring.find_class(class_id)
    assert p_cls is Person

    man2 = Person.deserialize_object(Person.s_bclass, None, ss)
    print(man == man2)

    d1 = Dict(
        _replace    = True,
        payload     = {1: [1,1], 2: [2,2]}
    )
    s1 = d1.serialize_object()

    d2 = Dict.deserialize_object(Dict.s_bclass, None, s1)
    print(d1 == d2)

    d3 = Dict(
        _replace = True,
        payload = {
            woman:      1.0,
            daughter:   2.0,
            son:        -1.0
        }
    )
    s3 = d3.serialize_object()

    d4 = Dict.deserialize_object(Dict.s_bclass, None, s3)
    print(d3 == d4)



from typing import Any

from core_10x.traitable import RT, Traitable

from ui_10x.traitable_editor import TraitableEditor, TraitableView
from ui_10x.utils import ux, ux_answer, ux_push_button, ux_success, ux_warning


class StockerPlug(Traitable):
    master: Traitable
    # fmt: off
    current_class_trait_name: str   = RT('current_class')
    current_entity_trait_name: str  = RT('current_entity')
    changed_entity_cb_name: str     = RT('on_changed_entity')        #-- f(ce: Traitable) - after accepting edits and reloading
    deleted_entity_cb_name: str     = RT('on_deleted_entity')        #-- f(de: Traitable)
    # fmt: on

    new_entity_cb: Any
    changed_entity_cb: Any
    deleted_entity_cb: Any

    current_class: type
    current_entity: Traitable

    def _cb(self, cb_name: str):
        m = self.master
        if m:
            cls = m.__class__
            method = getattr(cls, cb_name, None)
            if method:
                return lambda e: method(m, e)

        return lambda e: None

    # fmt: off
    def changed_entity_cb_get(self):    return self._cb(self.changed_entity_cb_name)
    def deleted_entity_cb_get(self):    return self._cb(self.deleted_entity_cb_name)
    # fmt: on

    def current_class_get(self) -> type:
        m = self.master
        if not m:
            return None

        cls = m.get_value(self.current_class_trait_name)
        if not cls:
            ce = self.current_entity
            if ce:
                cls = ce.__class__

        return cls

    def current_entity_get(self) -> Traitable:
        m = self.master
        if not m:
            return None

        return m.get_value(self.current_entity_trait_name)


class EntityStocker(Traitable):
    plug: StockerPlug
    entity_viewer: TraitableEditor
    buttons_spec: dict

    def entity_viewer_get(self) -> TraitableEditor:
        ce = self.plug.current_entity
        if not ce:
            return None

        return TraitableEditor.editor(ce, view=TraitableView.default(ce.__class__, read_only=True))

    def top_layout(self) -> ux.HBoxLayout:
        lay = ux.HBoxLayout()
        lay.set_spacing(0)

        for name, (cb, icon) in self.buttons_spec.items():
            lay.add_widget(ux_push_button(name, callback=cb, style_icon=icon))

        return lay

    def main_widget(self) -> ux.Widget:
        entity_viewer = self.entity_viewer
        if entity_viewer:
            lay = ux.VBoxLayout()
            lay.add_layout(self.top_layout(), stretch=0)
            lay.add_widget(entity_viewer.main_widget())

        else:
            lay = self.top_layout()

        w = ux.Widget()
        w.set_layout(lay)
        return w

    # def on_new_entity(self):
    #     cls = self.plug.current_class
    #     if cls:
    #         new_entity = cls()
    #         ed = TraitableEditor.editor(new_entity)
    #         if ed.popup(copy_entity = False, title = f'New Entity of {cls.__name__}', save = True):
    #             rc = new_entity.share(False)    #-- not accepting existing entity values, if any
    #             if not rc:
    #                 ux_warning(rc.error(), parent = None)   #-- TODO: parent must be an existing appropriate widget
    #                 #-- TODO: what to do with a temp new_entity?
    #
    #             else:
    #                 self.plug.new_entity_cb(new_entity)

    def on_edit_entity(self):
        ce = self.plug.current_entity
        if ce:
            ed = TraitableEditor.editor(ce)
            if ed.popup():
                self.plug.changed_entity_cb(ce)

    def on_reload_entity(self):
        ce = self.plug.current_entity
        if ce:
            if not ce.reload():
                ux_warning(f'Failed to reload entity {ce.__class__}/{ce.id()}', parent=None, on_close=lambda ctx: None)
            else:
                self.plug.changed_entity_cb(ce)

    def on_save_entity(self):
        ce = self.plug.current_entity
        if ce:
            try:
                rc = ce.save()
                if not rc:
                    ux_warning(rc.error(), parent=None, on_close=lambda ctx: None)

                return

            except Exception:  # -- revision conflict?
                # -- TODO: resolve revision conflict - MergingEditor
                ...

            # -- The conflict seems to be resolved, try again
            try:
                rc = ce.save()
                if not rc:
                    ux_warning(rc.error(), parent=None, on_close=lambda ctx: None)
                    return

            except Exception:
                ux_warning('Failed to resolve revision conflict (most probably due to continuous updates from other session(s)')
                return

            ux_success(f'Conflict resolved, {ce.__class__}/{ce.id()} has been saved', on_close=lambda ctx: None)

    def on_delete_entity(self):
        ce = self.plug.current_entity
        if ce:

            def on_close(accepted: bool):
                if accepted:
                    if not ce.delete():
                        ux_warning('Deletion failed: {ce.__class__}/{ce.id()', parent=None, on_close=lambda ctx: None)
                    self.plug.deleted_entity_cb(ce)

            ux_answer(f'Please confirm deletion of {ce.__class__}/{ce.id()}', parent=None, on_close=on_close)

    def buttons_spec_get(self) -> dict:
        # fmt: off
        return dict(
            #new     = (self.on_new_entity,      'FileIcon'),
            edit    = (self.on_edit_entity,     'FileDialogDetailedView'),
            reload  = (self.on_reload_entity,   'ArrowDown'),
            save    = (self.on_save_entity,     'DriveNetIcon'),
            delete  = (self.on_delete_entity,   'DialogDiscardButton')
        )
        # fmt: on



from core_10x.traitable import RT, M, Traitable


class A(Traitable):
    funny_message: str  = RT()

    def funny_message_get(self):
        dt = self.T.funny_message.data_type
        return f'{dt}'

class B(A):
    funny_message:int   = M()

class C(A):
    funny_message:float   = M()

    def funny_message_get(self):
        dt = self.T.funny_message.data_type
        return f'{dt}'

if __name__ == '__main__':
    a = A()
    b = B()
    c = C()

    print(f'A: {a.funny_message}; B: {b.funny_message}; C: {c.funny_message}')



from __future__ import annotations

from datetime import date, datetime

import pytest
from infra_10x.mongodb_store import MongoStore
from typing_extensions import Self

from core_10x.code_samples.person import Person
from core_10x.exec_control import BTP, CACHE_ONLY, GRAPH_OFF, GRAPH_ON, INTERACTIVE
from core_10x.rc import RC, RC_TRUE
from core_10x.traitable import AsOfContext, T


class MarriedPerson(Person):
    spouse: Self = T()

    def spouse_set(self, trait, spouse) -> RC:
        self.raw_set_value(trait, spouse)
        if spouse:
            spouse.raw_set_value(trait, self)
        return RC_TRUE


if __name__ == '__main__':
    with MongoStore.instance(hostname='localhost', dbname='traitable_history_test'):
        MarriedPerson.delete_collection()
        MarriedPerson.s_history_class.delete_collection()

        with BTP.create_root():
            # Create and save a person
            person = MarriedPerson(first_name='Alyssa', last_name='Lees', dob=date(1985, 7, 5), _replace=True)
            person.spouse = MarriedPerson(first_name='James', last_name='Bond', dob=date(1985, 7, 5), _replace=True)
            person.spouse.save()
            person.save()

            ts = datetime.utcnow()

            # Update and save again
            person.dob = date(1985, 7, 6)
            person.spouse.dob = date(1985, 7, 6)
            person.save(save_references=True).throw()

            person_id = person.id()
            print(person.serialize_object())

        with CACHE_ONLY():
            assert not MarriedPerson.existing_instance_by_id(person_id, _throw=False)

        for ctx in (
            INTERACTIVE,
            GRAPH_ON,
            GRAPH_OFF,
        ):
            for as_of in (
                ts,
                None,
            ):
                with ctx():
                    person1 = MarriedPerson(person_id)
                    assert person1.dob == date(1985, 7, 6)

                    person_as_of = MarriedPerson.as_of(person_id, as_of_time=ts)

                    with AsOfContext([MarriedPerson], as_of_time=as_of):
                        with pytest.raises(RuntimeError, match=r'MarriedPerson/Alyssa|Lees: object not usable - origin cache is not reachable'):
                            _ = person1.dob

                        with pytest.raises(RuntimeError, match=r'MarriedPerson/Alyssa|Lees: object not usable - origin cache is not reachable'):
                            _ = person_as_of.dob

                        person2 = MarriedPerson(person_id)
                        assert person2.dob == (date(1985, 7, 5 + (as_of is None)))
                        assert person2.spouse.dob == (date(1985, 7, 5 + (as_of is None)))

                    assert person1.spouse.dob == date(1985, 7, 6)
                    assert person_as_of.dob == date(1985, 7, 5)

                    assert person_as_of.spouse.dob == date(1985, 7, 6)  # note - since no AsOfContext was used, nested objects are not loaded "as_of".

                    with pytest.raises(RuntimeError, match=r'MarriedPerson/Alyssa|Lees: object not usable - origin cache is not reachable'):
                        _ = person2.dob



import rio


class LabeledCheckBox(rio.Component):
    label: str = ''
    is_on: bool = False
    is_sensitive: bool = True
    on_change: rio.EventHandler[rio.CheckboxChangeEvent] = None

    def build(self):
        return rio.Row(
            rio.Text(self.label),
            rio.Checkbox(
                is_on=self.bind().is_on,
                is_sensitive=self.is_sensitive,
                on_change=self.on_change,
            ),
        )



from __future__ import annotations

import os

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.serialization import (
    load_pem_private_key,
    load_pem_public_key,
)

from core_10x.global_cache import cache

# fmt: off
ROOT                = '.xx'
PUBLIC_EXP          = 65537
KEY_SIZE            = 2048
PRIVATE_KEY_FILE    = 'private.pem'
PUBLIC_KEY_FILE     = 'public.pem'
ENCODING            = 'utf-8'
# fmt: on


class SecKeys:
    @classmethod
    @cache
    def home_dir(cls) -> str:
        dir = f'{os.path.expanduser("~")}/{ROOT}'
        if os.path.exists(dir):
            assert os.path.isdir(dir), f'{dir} is not a directory'
        else:
            os.mkdir(dir)

        return dir

    @classmethod
    def generate_keys(cls, pwd=None) -> tuple:
        private_key = rsa.generate_private_key(public_exponent=PUBLIC_EXP, key_size=KEY_SIZE, backend=default_backend())
        public_key = private_key.public_key()

        if pwd:
            format = serialization.PrivateFormat.PKCS8
            algo = serialization.BestAvailableEncryption(bytes(pwd, encoding=ENCODING))
        else:
            format = serialization.PrivateFormat.TraditionalOpenSSL
            algo = serialization.NoEncryption()

        private_key_pem = private_key.private_bytes(encoding=serialization.Encoding.PEM, format=format, encryption_algorithm=algo)
        public_key_pem = public_key.public_bytes(encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo)

        return (private_key_pem, public_key_pem)

    @classmethod
    def create_keys(cls, keys: tuple = None) -> tuple:
        private_key, public_key = keys if keys else cls.generate_keys()
        home_dir = cls.home_dir()
        try:
            with open(f'{home_dir}/{PRIVATE_KEY_FILE}', 'wb') as f:
                f.write(private_key)
                os.chmod(f.name, 0o600)

            with open(f'{home_dir}/{PUBLIC_KEY_FILE}', 'wb') as f:
                f.write(public_key)
                os.chmod(f.name, 0o600)

            cls.keys.clear()  # -- cls.keys() will then read from the files
            return (private_key, public_key)

        except Exception:
            return (None, None)

    @classmethod
    @cache
    def keys(cls) -> tuple:
        home_dir = cls.home_dir()
        file = f'{home_dir}/{PRIVATE_KEY_FILE}'
        if os.path.exists(file):
            try:
                with open(file, 'rb') as f:
                    private_key = f.read()

                with open(f'{home_dir}/{PUBLIC_KEY_FILE}', 'rb') as f:
                    public_key = f.read()

                return (private_key, public_key)

            except Exception:
                pass

        return (None, None)

    @classmethod
    def encrypt(cls, message, public_key: bytes = None) -> bytes | None:
        if type(message) is str:
            message = bytes(message, encoding=ENCODING)

        if public_key is None:
            _, public_key = cls.keys()
            if public_key is None:
                return None

        public_key = load_pem_public_key(public_key)
        # fmt: off
        return public_key.encrypt(
            message,
            padding.OAEP(
                mgf         = padding.MGF1(algorithm = hashes.SHA256()),
                algorithm   = hashes.SHA256(),
                label       = None
            )
        )
        # fmt: on

    @classmethod
    def decrypt(cls, encrypted_message: bytes, to_str=True):
        private_key, _ = cls.keys()
        if private_key is None:
            return None

        private_key = load_pem_private_key(private_key, password=None)
        # fmt: off
        res = private_key.decrypt(
            encrypted_message,
            padding.OAEP(
                mgf         = padding.MGF1(algorithm = hashes.SHA256()),
                algorithm   = hashes.SHA256(),
                label       = None
            )
        )
        # fmt: on

        if to_str:
            res = res.decode(encoding=ENCODING)

        return res

    @classmethod
    def encrypt_private_key(cls, password) -> bytes | None:
        if type(password) is str:
            password = bytes(password, encoding=ENCODING)

        private_key, _ = cls.keys()
        if private_key is None:
            return None

        private_key = load_pem_private_key(private_key, password=None)
        # fmt: off
        return private_key.private_bytes(
            encoding                = serialization.Encoding.PEM,
            format                  = serialization.PrivateFormat.PKCS8,
            encryption_algorithm    = serialization.BestAvailableEncryption(password),
        )
        # fmt: on

    @classmethod
    def decrypt_private_key(cls, private_key_with_password, password) -> bytes:
        if type(password) is str:
            password = bytes(password, encoding=ENCODING)

        pk = load_pem_private_key(private_key_with_password, password=password)
        # fmt: off
        return pk.private_bytes(
            encoding                = serialization.Encoding.PEM,
            format                  = serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm    = serialization.NoEncryption(),
        )
        # fmt: on



from __future__ import annotations

import inspect
from typing import Any

from core_10x.py_class import PyClass
from core_10x.traitable import Traitable
from core_10x.xnone import XNone


class Directory:
    @staticmethod
    def _dir_class(value):
        if isinstance(value, Traitable):
            return DxTraitable

        if inspect.isclass(value) and issubclass(value, Traitable):
            return DxClass

        return Directory

    @staticmethod
    def instance(value: Any = None, **kwargs) -> Directory:
        dir_class = Directory._dir_class(value)
        return dir_class(value=value, **kwargs)

    # fmt: off
    def __init__(
        self,
        name: str           = '',
        value               = None,
        members: dict       = None,
        parent: Directory   = None
    ):
        # fmt: on
        self.check_dir_value(value)

        self.name = name
        self.value = value
        self.members = members or {}
        self.parent = parent

    @classmethod
    def check_dir_value(cls, value):
        f = getattr(value, '__hash__', None)
        assert f, f'Directory value must be hashable ({value})'

    def show_value(self) -> str:
        return str(self.value) if self.value else self.name

    def is_leaf(self) -> bool:
        return not self.members

    def subdirs(self) -> dict:
        return self.members

    def _add_subdir(self, subdir: Directory):
        self.members[subdir.value] = subdir
        subdir.parent = self

    def _get_or_add_subdir(self, value, **subdir_values):
        subdir = self.members.get(value)
        if subdir is None:
            subdir = Directory.instance(value, **subdir_values)
            subdir.parent = self
            self.members[value] = subdir

        return subdir

    def insert(self, value, *path, **subdir_values):
        if not path:
            self._get_or_add_subdir(value, **subdir_values)
        else:
            p_value = path[0]
            subdir = self._get_or_add_subdir(p_value)
            subdir.insert(value, *path[1:], **subdir_values)

    def insert_many(self, values: list, **subdir_values):
        for value in values:
            self._get_or_add_subdir(value, **subdir_values)

    def subdir_at(self, *path) -> Directory:
        if not path:
            return self

        subdir = self.members.get(path[0])
        return subdir.subdir_at(*path[1:]) if subdir else None

    def remove(self, value, *path) -> bool:
        subdir = self.subdir_at(*path)
        if not subdir:
            return False

        return self.members.pop(value, XNone) is not XNone

    def remove_everywhere(self, value):
        self.remove(value)
        for subdir in self.members.values():
            subdir.remove(value)

    def _find_paths(self, value, paths_found: list, *current_path):
        found = self.members.get(value)
        if found:
            paths_found.append((*current_path, value))

        for sub_value, subdir in self.members.items():
            subdir._find_paths(value, paths_found, *current_path, sub_value)

    def find_paths(self, value) -> list:
        paths_found = [()] if value == self.value else []
        self._find_paths(value, paths_found)
        return paths_found

    def _flatten(self, res: dict, *current_path):
        path = *current_path, self.value
        label = self.name
        if not label:
            label = self.show_value()
        res[path] = label
        for subdir in self.members.values():
            subdir._flatten(res, *path)

    def flatten(self, with_root: bool = False) -> dict:
        res = {}
        if with_root:
            self._flatten(res)
        else:
            for subdir in self.members.values():
                subdir._flatten(res)

        return res

    def choices(self, with_root: bool = False, path_delimiter: str = '/') -> dict:
        f = self.flatten(with_root = with_root)
        v_n_map = { path[-1]: name for path, name in f.items() }
        res = {}
        cr = 1
        for path in f.keys():
            label = path_delimiter.join(v_n_map[p] for p in path)
            value = path[-1]
            already_in = res.get(label)
            if already_in:
                if already_in == value:
                    continue

                label = f'{label}({cr})'
                cr += 1

            res[label] = value

        return res

    def contains(self, value) -> bool:
        if self.value == value:
            return True

        for subdir in self.members.values():
            if subdir.contains(value):
                return True

        return False

    def is_value_contained(self, value, by_other_value) -> bool:
        if self.value == by_other_value:
            return self.contains(value)

        for subdir in self.members.values():
            if subdir.is_value_contained(value, by_other_value):
                return True

        return False

    def _collect_leaf_values(self, res: set, data_accessor):
        if self.is_leaf():
            res.add(data_accessor(self.value))
        else:
            for subdir in self.members.values():
                subdir._collect_leaf_values(res, data_accessor)

    def leaf_values(self, data_accessor):       #-- data_accessor: callable(leaf_value)
        res = set()
        self._collect_leaf_values(res, data_accessor)
        return list(res)

    @classmethod
    def define(cls, value_name, members: list) -> Directory:
        if isinstance(value_name, tuple):     #-- ( value, name ) is given
            assert len(value_name) == 2, 'a pair of ( value, name ) is expected'
            value, name = value_name

        else:       #-- just a value
            value = value_name
            name = None

        dir = Directory.instance(value, name = name)

        sub_value_name = None
        for elem in members:
            if not sub_value_name:
                sub_value_name = elem
                continue

            if isinstance(elem, list):  #-- sub dirs are given
                sudir = cls.define(sub_value_name, elem)
                dir._add_subdir(sudir)
                sub_value_name = None

            else:   #-- no sub dirs are given, just value_name
                sudir = cls.define(sub_value_name, [])
                dir._add_subdir(sudir)
                sub_value_name = elem

        if sub_value_name:
            sudir = cls.define(sub_value_name, [])
            dir._add_subdir(sudir)

        return dir


class DxTraitable(Directory):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.value: Traitable
        self.name = self.value.id()

    @classmethod
    def check_dir_value(cls, value):
        assert isinstance(value, Traitable), f'Traitable is expected ({value})'

    def show_value(self) -> str:
        return self.value.id()  # -- TODO: id_repr()


class DxClass(Directory):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        cls = self.value
        self.name = PyClass.name(cls) if cls else ''

    @classmethod
    def check_dir_value(cls, value):
        assert inspect.isclass(value) and issubclass(value, Traitable), f'subclass of Traitable is expected ({value})'

    # def matching_traitables(self) -> dict:
    #     """
    #     Applicable for a Traitable class with a "reasonable" number of traitables in the collection
    #     TODO: need to provide a different mechanism (a view) for huge collections
    #     """
    #     cls = self.value
    #     return { e.id(): e for e in cls.shared_instances( _cache = self.cache_only(), **self.filters() ) }



from datetime import date

from core_10x.named_constant import NamedConstant
from core_10x.traitable import RC, RC_TRUE, RT, T, Traitable, Ui


class WEIGHT_QU(NamedConstant):
    LB =    ('lb',  1.)
    KG =    ('kg',  2.205)
    G =     ('g',   0.002205)
    CT =    ('ct',  0.0004409)

def verify_alpha(t, value: str) -> RC:
    if value.isalpha():
        return RC_TRUE
    return RC(False, f'{t.name} is {value}, but may have letters only')

class Person(Traitable):
    first_name: str         = T(T.ID)
    last_name: str          = T(T.ID)
    dob: date               = T()
    weight_lbs: float       = T(fmt = ',.4f')

    age: int
    full_name: str          = RT(T.EXPENSIVE)
    weight: float
    weight_qu: WEIGHT_QU    = RT(default = WEIGHT_QU.LB)

    older_than: bool

    def first_name_verify(self, t, value: str) -> RC:
        return verify_alpha(t, value)

    def last_name_verify(self, t, value: str) -> RC:
        return verify_alpha(t, value)

    def dob_set(self, trait, value: date) -> RC:
        today = date.today()
        if value and value > today:
            return RC(False, f'{value} may not be after {today}')
        return self.raw_set_value(trait, value)

    def full_name_get(self) -> str:
        return f'{self.first_name} {self.last_name}'

    def full_name_style_sheet(self) -> dict:
        age = self.age
        return {Ui.FG_COLOR: 'red' if age < 20 or age > 70 else 'green'}

    def full_name_set(self, trait, value: str) -> RC:
        self.print('full_name')
        parts = value.split(' ')
        if(len(parts) != 2):
            return RC(False, f'"{value}" - must be "first_name last_name"')

        self.first_name = parts[0]
        self.last_name = parts[1]
        return RC_TRUE

    def age_get(self) -> int:
        self.print('age')
        dob = self.dob
        if not dob:
            return 0

        today = date.today()
        years = today.year - dob.year
        dm = today.month - dob.month
        if dm < 0:
            return years - 1

        if dm > 0:
            return years

        return years - 1 if today.day < dob.day else years

    def older_than_get(self, age: int) -> bool:
        return self.age > age

    #-- getter and setter for weight trait
    def weight_get(self) -> float:
        self.print('weight')
        return self.weight_lbs / self.weight_qu.value

    def weight_set(self, trait, value) -> RC:  #-- trait setter gets its trait and the value and must return RC
        return self.set_values( weight_lbs = value * self.weight_qu.value)

    def weight_to_str(self, trait, value) -> str:
        return f'{trait.to_str(value)} {self.weight_qu.label}'

    def post_verify(self) -> RC:
        if self.age > 60 and self.weight_lbs > 200.:
            return RC(False, f'{self.full_name} - overweight condition detected')
        return RC_TRUE

    s_print = False
    def print(self, trait_name: str):
        cls = self.__class__
        if cls.s_print:
            print(f'inside {cls.__name__}.{trait_name}_get()')



from __future__ import annotations

from core_10x.ts_union import TsUnion
from ui_10x.examples.style_sheet import StyleSheet
from ui_10x.rio.component_builder import UserSessionContext
from ui_10x.traitable_editor import TraitableEditor

import rio


@rio.page(
    name='StyleSheet',
    url_segment='ss',
)
class StyleSheetPage(rio.Component):
    def build(self) -> rio.Component:
        session_context = self.session[UserSessionContext]
        if not session_context.traitable_store:
            session_context.traitable_store = TsUnion()
        with session_context.traitable_store:
            return TraitableEditor(StyleSheet(), _confirm=True).dialog().build(self.session)



"""Unit tests for TsStore abstract bases. Transaction tests run from testlib against ts_instance (see conftest)."""

import pytest

from core_10x.testlib.fixtures import with_transactions
from core_10x.ts_store import TsCollection, TsStore
from core_10x.ts_union import TsUnion

from core_10x.testlib.ts_store_transaction_tests import TestSaveIfChanged, TestTsStoreTransaction  # collected by pytest
from core_10x.testlib.ts_tests import TestTSStore, ts_setup # collected by pytest


class TestAbstractBases:
    def test_ts_store_is_abstract(self):
        with pytest.raises(TypeError, match="abstract"):
            # noinspection PyAbstractClass
            TsStore()

    def test_ts_collection_is_abstract(self):
        with pytest.raises(TypeError, match="abstract"):
            # noinspection PyAbstractClass
            TsCollection()


    def test_ts_union_no_stores_returns_no_op_transaction(self):
        union = TsUnion()
        with union.transaction() as tx:
            assert not tx.ended
        assert tx.ended



import math
import sys

class PInfType:
    def __hash__(self):                         return id(self)

    def __int__(self):                          return sys.maxsize
    def __bool__(self):                         return True
    def __float__(self):                        return math.inf
    def __repr__(self):                          return chr(8734)

    def __invert__(self):                       return MInf

    def __eq__(self, other):                    return other is self
    def __ne__(self, other):                    return not other is self
    def __lt__(self, other):                    return False
    def __le__(self, other):                    return other is self
    def __gt__(self, other):                    return not other is self
    def __ge__(self, other):                    return True

    def __abs__(self):                          return self
    def __neg__(self):                          return MInf
    def __add__(self, other):                   return self
    def __sub__(self, other):                   return self     #-- XInf - XInf is XInf (not NaN)
    def __mul__(self, other):                   return self
    def __pow__(self, power, modulo = None):    return self
    def __truediv__(self, other):               return self
    def __floordiv__(self, other):              return self

    def __init_subclass__(cls, **kwargs):
        raise TypeError('May not derive from Inf')

class MInfType:
    def __hash__(self):                         return id(self)

    def __int__(self):                          return ~sys.maxsize
    def __float__(self):                        return -math.inf
    def __repr__(self):                         return f'-{PInf.__repr__()}'

    def __invert__(self):                       return PInf

    def __eq__(self, other):                    return other is self
    def __ne__(self, other):                    return not other is self
    def __lt__(self, other):                    return not other is self
    def __le__(self, other):                    return True
    def __gt__(self, other):                    return False
    def __ge__(self, other):                    return other is self

    def __abs__(self):                          return PInf
    def __neg__(self):                          return PInf
    def __add__(self, other):                   return self
    def __sub__(self, other):                   return self
    def __mul__(self, other):                   return self
    def __pow__(self, power, modulo = None):    return self
    def __truediv__(self, other):               return self
    def __floordiv__(self, other):              return self

    def __init_subclass__(cls, **kwargs):
        raise TypeError('May not derive from Inf')

MInf = MInfType()
PInf = PInfType()
XInf = PInf




from core_10x.nucleus import Nucleus

class MongoCollectionHelperStub:
    @classmethod
    def prepare_filter_and_pipeline(cls, serialized_traitable, filter, pipeline):
        rev_tag = Nucleus.REVISION_TAG()
        id_tag = '_id'
        for key in (rev_tag, id_tag):
            filter[key] = serialized_traitable.pop(key)

        rev_condition = {'$and': [{'$eq': ['$' + name, {'$literal': value}]} for name, value in serialized_traitable.items()]}

        # fmt: off
        update_revision = {
            '$cond': [
                rev_condition,         #-- if each field is equal to its prev value
                filter[rev_tag],       #       then, keep the revision as is
                filter[rev_tag] + 1    #       else, increment it
            ]
        }

        pipeline.append(
            {
                '$replaceRoot': {
                    'newRoot': {
                        id_tag:     filter[id_tag],
                        rev_tag:    update_revision,
                    }
                }
            }
        )
        # fmt: on

        pipeline.extend(
            {'$replaceWith': {'$setField': dict(field=field, input='$$ROOT', value={'$literal': value})}}
            for field, value in serialized_traitable.items()
        )



import ast
import inspect

from core_10x.named_constant import EnumBits, NamedConstant, NamedCallable
from core_10x.nucleus import Nucleus
from core_10x.package_refactoring import PackageRefactoring
from core_10x.py_class import PyClass
from core_10x.rc import RC
from core_10x.trait import T, Trait
from core_10x.ui_hint import Ui
from core_10x.xdate_time import XDateTime, date, datetime
from core_10x.xnone import XNone, XNoneType


class primitive_trait(Trait, register=False):
    s_ui_hint = Ui.NONE

    def default_value(self):
        return self.default

    def same_values(self, value1, value2) -> bool:
        return value1 == value2

    def from_str(self, s: str):
        try:
            v = ast.literal_eval(s)
            dt = self.data_type
            return v if type(v) is dt else dt(v)
        except Exception:
            return XNone

    def from_any_xstr(self, value):
        return self.data_type(value)

    # def to_str(self, v) -> str:
    #      return str(v)

    def to_id(self, v) -> str:
        return str(v)

    def serialize(self, value):
        return value

    def deserialize(self, value):
        return value


class bool_trait(primitive_trait, data_type=bool):
    s_ui_hint = Ui.check()

    # fmt: off
    FORMATS     = [
        ('yes',''),
        ('yes','no'),
        ('true','false'),
        ('on','off')
    ]
    s_fmt         = FORMATS[0]
    # fmt: on

    def to_id(self, value: bool) -> str:
        return '0' if value else '1'

    def from_str(self, s: str):
        res = super().from_str(s)
        if res is not XNone:
            return res
        s = s.strip().lower()
        return next((f[0] == s for f in self.FORMATS if s in f), XNone)


class int_trait(primitive_trait, data_type=int):
    s_ui_hint = Ui.line()

    s_fmt = ','


class float_trait(primitive_trait, data_type=float):
    s_ui_hint = Ui.line()

    s_fmt = ',.2f'

    # def from_str(self, s: str) -> RC:
    #    return RC(True, locale.atof(s))

    def is_acceptable_type(self, data_type: type) -> bool:
        return data_type is float or data_type is int


class str_trait(primitive_trait, data_type=str):
    s_ui_hint = Ui.line(align_h=-1)

    # pattern = ''
    # placeholder = ''

    def from_str(self, s: str):
        return s

    def to_str(self, v) -> str:
        return v

    def to_id(self, value: str) -> str:
        return value


# class bin_trait(Trait, data_type = bytes):
#     w = 0   #-- max width, if any and relevant
#     h = 0   #-- max height, if any and relevant
#
# def pixmap(*args, **kwargs):
#     tdef: TraitDefinition = T(bytes, *args, **kwargs)
#     tdef.set_widget_type(Ui.WIDGET_TYPE.PIXMAP)
#     return tdef


class datetime_trait(Trait, data_type=datetime):
    s_ui_hint = Ui.line(align_h=0)

    def from_str(self, s: str):
        dt = XDateTime.str_to_datetime(s)
        if dt is None:
            return XNone
        return dt

    def from_any_xstr(self, value):
        dt = XDateTime.to_datetime(value)
        if dt is None:
            return XNone
        return dt

    def to_str(self, v: datetime) -> str:
        return XDateTime.datetime_to_str(v)

    # s_acceptable_types = { datetime, date, int, str }
    # def is_acceptable_type(self, data_type: type) -> bool:
    #     return data_type in self.s_acceptable_types

    # -- NOTES:
    # -- 1) we believe datetime is mostly acceptable for a storage, e.g., Mongo
    # -- 2) it is in UTC, so no locale's datetime issues

    def serialize(self, value: datetime):
        return value

    def deserialize(self, value: datetime):
        return value

    def to_id(self, value) -> str:
        return XDateTime.datetime_to_str(value, with_ms=True)


class date_trait(Trait, data_type=date):
    s_ui_hint = Ui.line(min_width=10, align_h=0)

    def from_str(self, s: str):
        dt = XDateTime.str_to_date(s)
        if dt is None:
            return XNone
        return dt

    def from_any_xstr(self, value):
        dt = XDateTime.to_date(value)
        if dt is None:
            return XNone
        return dt

    def to_str(self, v: date) -> str:
        return XDateTime.date_to_str(v)

    # s_acceptable_types = { datetime, date, int, str }
    # def is_acceptable_type(self, data_type: type) -> bool:
    #     return data_type in self.s_acceptable_types

    def serialize(self, value: date):
        return XDateTime.date_to_str(value, format=XDateTime.FORMAT_X10)

    def deserialize(self, value):
        return XDateTime.str_to_date(value, format=XDateTime.FORMAT_X10)

    def to_id(self, value) -> str:
        return XDateTime.date_to_str(value)


class bytes_trait(Trait, data_type=bytes):
    s_ui_hint = Ui.NONE

    def to_str(self, value):
        return value.decode(encoding='utf-8')

    def to_id(self, value) -> str:
        return self.to_str(value)

    def same_values(self, value1, value2) -> bool:
        return value1 == value2

    def from_str(self, s: str):
        return bytes(s, encoding='utf-8')

    def from_any_xstr(self, value):
        return XNone  # may not be called!

    # def is_acceptable_type(self, data_type: type) -> bool:
    #    ...

    def serialize(self, value):
        return value

    def deserialize(self, value):
        return value


class class_trait(Trait, data_type=type):
    s_ui_hint = Ui.line(align_h=-1)

    def to_str(self, value):
        return PyClass.name(value)

    def to_id(self, value) -> str:
        return PackageRefactoring.find_class_id(value)

    def same_values(self, value1, value2) -> bool:
        return value1 is value2

    def from_str(self, s: str):
        return PackageRefactoring.find_class(s)

    def from_any_xstr(self, value):
        return XNone  # may not be called!

    def is_acceptable_type(self, data_type: type) -> bool:
        return inspect.isclass(data_type)

    def serialize(self, value):
        return PackageRefactoring.find_class_id(value)

    def deserialize(self, value: str):
        return PackageRefactoring.find_class(value)


class list_trait(Trait, data_type=list):
    s_ui_hint = Ui.line()

    def to_str(self, v) -> str:
        return str(v)

    def to_id(self, value) -> str:
        return str(value)

    def default_value(self) -> list:
        return list(self.default)

    def serialize(self, value: list):
        return Nucleus.serialize_list(value, self.flags_on(T.EMBEDDED))

    def deserialize(self, value: list):
        return Nucleus.deserialize_list(value)

class set_trait(list_trait, data_type = set):
    def default_value(self) -> set:
        return set(self.default)

    def serialize(self, value: set):
        return Nucleus.serialize_list(value, self.flags_on(T.EMBEDDED))

    def deserialize(self, value: list):
        return set(Nucleus.deserialize_list(value))

class dict_trait(Trait, data_type=dict):
    s_ui_hint = Ui.line(flags=Ui.SELECT_ONLY)

    def use_format_str(self, fmt: str, value) -> str:
        return str(value) if not fmt else fmt.join(f"'{key}' -> '{val}'" for key, val in value.items())

    def to_str(self, v) -> str:
        return str(v)

    def to_id(self, value) -> str:
        return str(value)

    def default_value(self) -> dict:
        return dict(self.default)

    def serialize(self, value: dict):
        return Nucleus.serialize_dict(value, self.flags_on(T.EMBEDDED))

    def deserialize(self, value: dict):
        return Nucleus.deserialize_dict(value)


class any_trait(Trait, data_type=XNoneType):  # -- any
    s_ui_hint = Ui.NONE

    def to_str(self, v) -> str:
        return str(v)

    def to_id(self, value) -> str:
        return str(value)

    def from_str(self, s: str):
        return None

    def from_any_xstr(self, value):
        return value

    def is_acceptable_type(self, data_type: type) -> bool:
        return True

    def serialize(self, value):
        return Nucleus.serialize_any(value, self.flags_on(T.EMBEDDED))

    def deserialize(self, value):
        return Nucleus.deserialize_any(value)

    def same_values(self, value1, value2) -> bool:
        return value1 is value2


class nucleus_trait(Trait, data_type=Nucleus, base_class=True):
    s_ui_hint = Ui.NONE

    def to_str(self, v) -> str:
        return self.data_type.to_str(v)

    def to_id(self, v) -> str:
        return self.data_type.to_id(v)

    def from_str(self, s: str) -> Nucleus:
        return self.data_type.from_str(s)

    def from_any_xstr(self, value) -> Nucleus:
        return self.data_type.from_any_xstr(value)

    def is_acceptable_type(self, data_type: type) -> bool:
        return issubclass(data_type, self.data_type)

    def same_values(self, value1, value2) -> bool:
        return self.data_type.same_values(value1, value2)

    def serialize(self, value: Nucleus):
        if type(value) is self.data_type:
            return value.serialize(self.flags_on(T.EMBEDDED))

        return Nucleus.serialize_any(value, self.flags_on(T.EMBEDDED))

    def deserialize(self, serialized_value) -> Nucleus:
        if isinstance(serialized_value, dict) and (value := Nucleus.deserialize_record(serialized_value)):
            return value

        return self.data_type.deserialize(serialized_value)

    def choices(self):
        return self.data_type.choose_from()


class named_constant_trait(nucleus_trait, data_type=NamedConstant, base_class=True):
    s_ui_hint = Ui.choice(flags=Ui.SELECT_ONLY)

    def is_acceptable_type(self, data_type: type) -> bool:
        return data_type is self.data_type

class named_callable_trait(nucleus_trait, data_type = NamedCallable, base_class = True):
    s_ui_hint = Ui.choice(flags = Ui.SELECT_ONLY)

    def is_acceptable_type(self, data_type: type) -> bool:
        return data_type is self.data_type

    def serialize(self, value: NamedCallable):
        if value.name is None:
            raise TypeError(f"Trait {self.name} = {value} - is just a callable, but supposed to be a NamedCallable")

        return super().serialize(value)

    def default_setter(obj, trait: Trait, value: NamedCallable) -> RC:
        if not isinstance(value, NamedCallable):
            if not callable(value):
                raise TypeError(f'{obj.__class__}/{obj.id()}.{trait.name}: attempt to set a non-callable {value}')
            value = NamedCallable.just_func(value)

        return obj.raw_set_value(trait, value)

    def post_ctor(self):
        if not self.f_set:
            self.set_f_set(self.__class__.default_setter, False)

class flags_trait(nucleus_trait, data_type=EnumBits, base_class=True):
    s_ui_hint = Ui.line()

    def is_acceptable_type(self, data_type: type) -> bool:
        return data_type is self.data_type



if __name__ == '__main__':
    from core_10x.code_samples.person import Person
    from core_10x.exec_control import INTERACTIVE
    from core_10x.ts_store import TsStore

    from ui_10x.entity_stocker import EntityStocker, StockerPlug
    from ui_10x.utils import UxDialog, ux

    ux.init()

    with TsStore.instance_from_uri('mongodb://localhost/test'):
        plug = StockerPlug(current_class = Person)
        se = EntityStocker(plug = plug)

        with INTERACTIVE() as graph:
            w = se.main_widget()
            d = UxDialog(w)
            rc = d.exec()



from __future__ import annotations

import inspect
from functools import partial
from typing import TYPE_CHECKING

import rio
import ui_10x.platform_interface as i
from ui_10x.rio.component_builder import Widget

if TYPE_CHECKING:
    from collections.abc import Callable


class Label(Widget, i.Label):
    s_component_class = rio.Text
    s_forced_kwargs = {'selectable': False}  # , 'align_x': 0}
    s_single_child = True
    s_children_attr = 'text'
    s_default_kwargs = {'text': ''}


class PushButton(Label, i.PushButton):
    s_component_class = rio.Button
    s_forced_kwargs = {}
    s_children_attr = 'content'
    s_default_kwargs = {'content': ''}

    def clicked_connect(self, bound_method: Callable[[bool], None]):
        unbound_params = [
            p.name for p in inspect.signature(bound_method).parameters.values() if p.default is p.empty and not p.kind.name.startswith('VAR_')
        ]
        assert len(unbound_params) < 2, f'Expected 0 or 1 unbound parameters in {bound_method.__name__}, but found {unbound_params}'
        if len(unbound_params) > 0:
            bound_method = partial(bound_method, False)
        self['on_press'] = self.callback(bound_method)

    def set_flat(self, flat: bool):
        self['style'] = 'plain-text' if flat else 'major'

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        if args:
            label = args[-1]
            assert label is not None
            icon = args[0] if len(args) > 1 else None
            self.set_text(label)
            self['icon'] = icon



from core_10x.exec_control import CACHE_ONLY

if __name__ == '__main__':
    from core_10x.code_samples.person import Person
    from core_10x.exec_control import GRAPH_ON

    with (GRAPH_ON(convert_values = True, debug = True),CACHE_ONLY()):
        p = Person(first_name = 'John', last_name = 'Smith')

        print(p.full_name)
        p.dob = '19630514'
        p.weight_lbs = 170

        ser = p.serialize_object()

    print(p,ser)



from __future__ import annotations

import functools
from weakref import WeakKeyDictionary

import pytest
from core_10x.code_samples.person import WEIGHT_QU, Person
from core_10x.exec_control import BTP, CACHE_ONLY, GRAPH_OFF, GRAPH_ON, INTERACTIVE
from core_10x.trait_definition import RT, T
from core_10x.traitable import Traitable
from core_10x.traitable_id import ID
from core_10x.xnone import XNone


def call_counter(method):
    @functools.wraps(method)
    def ctr(self, *args, **kwargs):
        ctr.call_counts[self] = ctr.call_counts.get(self, 0) + 1
        return method(self, *args, **kwargs)

    ctr.call_counts = WeakKeyDictionary()
    ctr.call_count = lambda obj: ctr.call_counts.get(obj, 0) if obj else sum(ctr.call_counts.values())
    return ctr


class CallCount:
    def __init__(self, obj, method):
        self.obj = obj
        self.method = method
        assert hasattr(method, 'call_count')

    def __enter__(self):
        self.method.call_counts[self.obj] = 0
        return self

    def __exit__(self, *args):
        pass

    @property
    def call_count(self):
        return self.method.call_count(self.obj)


class TestablePerson(Person):
    older_than_get = call_counter(Person.older_than_get)
    weight_get = call_counter(Person.weight_get)
    weight_set = call_counter(Person.weight_set)

    young: bool = RT()

    @call_counter
    def young_get(self) -> bool:
        return self.age < 30

    @classmethod
    def load_data(cls, id: ID) -> dict | None:
        return None


@pytest.fixture
def testable_person():
    with CACHE_ONLY():
        return TestablePerson(first_name='John', last_name='Smith')


def reset_person(p):
    for trait in p.s_dir.values():
        if not trait.getter_params:
            p.invalidate_value(trait)
            if not trait.flags_on(T.ID):  # id traits may be set in base layer
                if trait.f_get.__name__ == 'default_value':  # TODO: replace with flags
                    assert p.get_value(trait) is trait.default, trait.name
                elif trait.f_get.__name__.endswith('_get'):
                    assert p.get_value(trait) == getattr(p, f'{trait.name}_get')()
                else:
                    assert p.get_value(trait) == trait.f_get(p)
        else:
            # TODO: use nodes?
            ...


def test_get_set(testable_person):
    on = BTP.current().flags() & BTP.ON_GRAPH
    p = testable_person
    p.weight_lbs = 100.0
    p.weight_qu = WEIGHT_QU.LB

    assert p.weight == 100

    with CallCount(p, TestablePerson.weight_get) as cc:
        assert p.weight == 100
        assert cc.call_count == (not on)

    p.weight_qu = WEIGHT_QU.KG
    assert p.weight_qu is WEIGHT_QU.KG

    with CallCount(p, TestablePerson.weight_get) as cc:
        assert p.weight == 100 / WEIGHT_QU.KG.value
        assert cc.call_count == 1

    # now use the setter...
    with CallCount(p, TestablePerson.weight_set) as cc:
        p.weight = 100.0
        assert cc.call_count == 1

    with CallCount(p, TestablePerson.weight_get) as cc:
        assert p.weight == 100
        assert cc.call_count == 1

    with CallCount(p, TestablePerson.weight_get) as cc:
        assert p.weight_lbs == 100 * WEIGHT_QU.KG.value
        assert cc.call_count == 0


def test_dep_change(testable_person):
    on = BTP.current().flags() & BTP.ON_GRAPH
    p = testable_person
    p.age = 30

    assert not p.young

    with CallCount(p, TestablePerson.young_get) as cc:
        assert not p.young
        assert cc.call_count == (not on)

    p.age = 20
    with CallCount(p, TestablePerson.young_get) as cc:
        assert p.young
        assert cc.call_count == 1

    p.age = 30


def test_dep_change_with_arg(testable_person):
    on = BTP.current().flags() & BTP.ON_GRAPH

    p = testable_person
    p.age = 30

    assert not p.older_than(30)

    with CallCount(p, TestablePerson.older_than_get) as cc:
        assert not p.older_than(30)
        assert cc.call_count == (not on)

    p.age = 40
    with CallCount(p, TestablePerson.older_than_get) as cc:
        assert p.older_than(30)
        assert cc.call_count == 1

    p.age = 30


def test_nested():
    class P(Traitable):
        first_name: str = RT()
        last_name: str = RT()
        full_name: str = RT()

        def full_name_get(self):
            return f'{self.first_name} {self.last_name}'

    p = P(first_name='Jane', last_name='Smith')
    assert p.full_name == 'Jane Smith'
    with INTERACTIVE() as i1:
        p.last_name = 'Baker'
        assert p.full_name == 'Jane Baker'
        with INTERACTIVE() as i2:
            p.first_name = 'Tom'
            assert p.full_name == 'Tom Baker'
        assert p.full_name == 'Jane Baker'
        with i2:
            assert p.full_name == 'Tom Baker'
        assert p.full_name == 'Jane Baker'

    assert p.full_name == 'Jane Smith'
    with i1:
        assert p.full_name == 'Jane Baker'

    assert p.full_name == 'Jane Smith'


def test_convert(testable_person, on=False):
    assert on == bool(BTP.current().flags() & BTP.CONVERT_VALUES)
    if not on:
        with pytest.raises(TypeError):
            Person(first_name=1, last_name=2)
    else:
        p = Person(first_name=1, last_name=2)
        assert p.full_name == '1 2'


def test_graph(testable_person, on=False):
    assert on == bool(BTP.current().flags() & BTP.ON_GRAPH)
    test_dep_change(testable_person)
    test_dep_change_with_arg(testable_person)
    test_get_set(testable_person)


def test_exec_control(testable_person, graph=False, convert=False, debug=False):
    with CACHE_ONLY():
        test_graph(testable_person, on=graph)
        reset_person(testable_person)
        test_convert(testable_person, on=convert)
        reset_person(testable_person)
        if not convert:
            test_debug(testable_person, on=debug)


def test_repro(testable_person):
    with GRAPH_ON():
        testable_person.weight_lbs = 100
        reset_person(testable_person)
        with GRAPH_OFF():
            testable_person.weight_lbs = 100
            reset_person(testable_person)


def test_graph_on(testable_person):
    with GRAPH_ON():
        test_exec_control(testable_person, True, False, False)
        with GRAPH_OFF():
            test_exec_control(testable_person, False, False, False)
            reset_person(testable_person)


def test_debug(testable_person, on=False):
    assert on == bool(BTP.current().flags() & BTP.DEBUG)
    p = TestablePerson(first_name='John', last_name='Smith')
    assert p.weight_lbs is XNone

    p.weight_lbs = 100.0

    if on:
        with pytest.raises(TypeError):
            p.weight_lbs = '200'

        assert p.weight == 100.0

    p.weight_lbs = 200
    assert p.weight == 200.0

    p.invalidate_value(p.T.weight_lbs)
    assert p.weight_lbs is XNone


def test_graph_convert_debug(testable_person):
    with GRAPH_ON(convert_values=True, debug=True):
        test_exec_control(testable_person, True, True, True)
        with GRAPH_OFF():
            test_exec_control(testable_person, False, True, True)
        with GRAPH_OFF(convert_values=False):
            test_exec_control(testable_person, False, False, True)
        with GRAPH_OFF(debug=False, convert_values=False):
            test_exec_control(testable_person, False, False, False)


def test_graph_convert(testable_person):
    with GRAPH_ON(convert_values=True):
        test_exec_control(testable_person, True, True, False)
        with GRAPH_OFF():
            test_exec_control(testable_person, False, True, False)
        with GRAPH_OFF(convert_values=False):
            test_exec_control(testable_person, False, False, False)
        with GRAPH_OFF(debug=True, convert_values=False):
            test_exec_control(testable_person, False, False, True)


def test_graph_debug(testable_person):
    with GRAPH_ON(debug=True):
        test_exec_control(testable_person, True, False, True)

        with GRAPH_OFF():
            test_exec_control(testable_person, False, False, True)
        with GRAPH_OFF(debug=False):
            test_exec_control(testable_person, False, False, False)
        with GRAPH_OFF(debug=False, convert_values=True):
            test_exec_control(testable_person, False, True, False)


@pytest.mark.parametrize('on_graph', [0, 1])
@pytest.mark.parametrize('debug', [0, 1])
@pytest.mark.parametrize('convert_values', [0, 1])
@pytest.mark.parametrize('on_graph2', [0, 1])
@pytest.mark.parametrize('debug2', [0, 1])
@pytest.mark.parametrize('convert_values2', [0, 1])
def test_traitable_processor_creation(on_graph, debug, convert_values, on_graph2, debug2, convert_values2):
    class X(Traitable):
        x: int = RT(T.ID)
        v: int = RT()

    X(x=1, v=10, _replace=True)
    assert X(x=1).v == 10

    default_cache = BTP.current().cache()
    with BTP.create(on_graph, debug, convert_values, use_parent_cache=False, use_default_cache=False):
        X(x=2, v=20, _replace=True)
        parent_cache = BTP.current().cache()
        assert default_cache is not parent_cache
        with BTP.create(on_graph2, debug2, convert_values2, use_parent_cache=False, use_default_cache=False):
            X(x=3, v=30, _replace=True)
            child_cache = BTP.current().cache()
            assert child_cache is not parent_cache
            assert child_cache is not default_cache
            assert X(x=1).v == 10
            assert X(x=2).v == 20

        with BTP.create(on_graph2, debug2, convert_values2, use_parent_cache=True, use_default_cache=False):
            X(x=4, v=40, _replace=True)
            child_cache = BTP.current().cache()
            if on_graph == on_graph2:  # use_parent_cache only works if both parent and child are either on_graph or off_graph
                assert child_cache is parent_cache
            else:
                assert child_cache is not parent_cache
                assert child_cache is not default_cache
        with BTP.create(on_graph2, debug2, convert_values2, use_parent_cache=False, use_default_cache=True):
            X(x=5, v=50, _replace=True)
            child_cache = BTP.current().cache()
            if not on_graph2:  # use_default_cache only works for off-graph mode
                assert child_cache is default_cache
                assert X(x=2).v is XNone
            else:
                assert child_cache is not parent_cache
                assert child_cache is not default_cache
                assert X(x=2).v == 20  # from parent
            assert X(x=1).v == 10

        assert X(x=1).v == 10  # from parent caches
        assert X(x=2).v == 20  # set in this cache
        assert X(x=3).v is XNone  # not set as was set in child with use_parent_cache=False
        if on_graph == on_graph2:  # use_parent_cache only works if both parent and child are either on_graph or off_graph
            assert X(x=4).v == 40  # this cache, as was set in child with use_parent_cache=True
        if not on_graph2:  # use_default_cache only works for off-graph mode
            assert X(x=5).v == 50  # in default cache, which is our parent

    assert X(x=1).v == 10
    assert X(x=2).v is XNone
    assert X(x=3).v is XNone
    assert X(x=4).v is XNone
    if not on_graph2:  # use_default_cache only works for off-graph mode
        assert X(x=5).v == 50  # we are in default cache, and it was set in default cache

    # TODO assert behavior with conflicting params, e.g.
    #
    # with pytest.raises(Exception):
    #     BTP.create(-1,-1,-1,use_parent_cache=True,use_default_cache=True)

    # TODO: node set, cleared, checked - each in nested cache level



import sys

import pytest


def test_platform_import_raises(monkeypatch):
    monkeypatch.delitem(sys.modules, 'ui_10x.platform', raising=False)
    with pytest.raises(ImportError):  # raises due to incorrect UI_PLATFORM from conftest
        import ui_10x.platform



from __future__ import annotations

from functools import partial

import rio


class RadioButton(rio.Component):
    label: str = ''
    value: str = ''
    checked: bool = False
    on_select: rio.EventHandler[[]] = None

    async def on_press(self, button) -> None:
        if self.on_select:
            self.on_select()
        else:
            self.checked = not self.checked
        button.icon = self.icon_name()

    def icon_name(self) -> str:
        return f'radio_button_{"checked" if self.checked else "unchecked"}'

    def build(self) -> rio.Component:
        # Use an icon to visually represent the radio button state
        icon_button = rio.IconButton(self.icon_name())
        icon_button.on_press = partial(self.on_press, icon_button)
        return rio.Row(
            icon_button,
            rio.Text(self.label),
            spacing=0.5,  # Space between icon and label
        )



from core_10x.code_samples.person import Person
from core_10x.exec_control import CACHE_ONLY
from core_10x.trait_filter import BETWEEN, NE, OR, f

if __name__ == '__main__':
    with CACHE_ONLY():
        p = Person(first_name = 'Sasha', last_name = 'Davidovich')

    r = OR(
        f(age=BETWEEN(50, 70), first_name=NE('Sasha')),
        f(age=17)
    )

    print(r.prefix_notation())
    print(r.eval(p))

    name = 'IBM Book'
    query = f(OR(f(book1_name = name), f(book2_name = name)))
    print(query.prefix_notation())



from contextlib import nullcontext

import pytest

from infra_10x import MongoCollectionHelper
from infra_10x.mongodb_store import MongoStore
from infra_10x.testlib.mongo_collection_helper import MongoCollectionHelperStub

MONGO_URL = 'mongodb://localhost:27017/'
# Example with X509 auth (replace with your URI; do not commit real hostnames or paths):
# MONGO_URL="mongodb+srv://HOST/?authMechanism=MONGODB-X509&authSource=%24external&tls=true&tlsCertificateKeyFile=/path/to/client.pem"
TEST_DB = 'test_db'


@pytest.fixture(params=[True, False], ids=['cxx-helper', 'py-helper'])
def ts_instance(mocker,request):
    instance = MongoStore.instance(hostname=MONGO_URL, dbname=TEST_DB)
    instance.username = 'test_user'
    if not instance.supports_transactions():
        instance.transaction = lambda *args: nullcontext()
    if not request.param:
        # Intentionally materialize metaclass __annotations__ to simulate prior
        # access on pybind11_type and keep this path covered.
        _ = type(MongoCollectionHelper).__annotations__
        mock = mocker.patch('infra_10x.MongoCollectionHelper')
        mock.value = MongoCollectionHelperStub
    return instance



from core_10x.exec_control import GRAPH_ON, BTP

class Scenario:
    s_instances = {}
    def __new__(cls, name: str = None, debug: bool = -1, convert_values: bool = -1):
        s = cls.s_instances.get(name)
        if s is None:
            s = object.__new__(cls)
            if name:
                cls.s_instances[name] = s
            s.name = name
            s.btp = GRAPH_ON(debug = debug, convert_values = convert_values)

        return s

    def __enter__(self):
        self.btp.begin_using()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.btp.end_using()
        if not self.name:
            self.btp = None


from .application import Application, Style
from .button import Label, PushButton
from .button_group import ButtonGroup, RadioButton
from .calendar import CalendarWidget
from .checkbox import CheckBox
from .dialog import Dialog
from .group_box import GroupBox
from .layout import FormLayout, HBoxLayout, VBoxLayout
from .line_edit import LineEdit
from .list import FindFlags, ListItem, ListWidget
from .message_box import MessageBox, Separator
from .scroll_area import SCROLL, ScrollArea
from .spacer import Spacer
from .splitter import Direction, Splitter
from .text_edit import TextEdit
from .tree import TreeItem, TreeWidget

__all__ = [
    'SCROLL',
    'Application',
    'ButtonGroup',
    'CalendarWidget',
    'CheckBox',
    'Dialog',
    'Direction',
    'FindFlags',
    'FormLayout',
    'GroupBox',
    'HBoxLayout',
    'Label',
    'LineEdit',
    'ListItem',
    'ListWidget',
    'MessageBox',
    'PushButton',
    'RadioButton',
    'ScrollArea',
    'Separator',
    'Spacer',
    'Splitter',
    'Style',
    'TextEdit',
    'TreeItem',
    'TreeWidget',
    'VBoxLayout',
]



import inspect

from core_10x.xnone import XNone

# ===================================================================================================================================
#   @cache must be used to cache results of function calls globally (at the process level)
#
#   Notes:
#
#   - non-hashable args are NOT supported for performance reasons
#   - args MUST follow the declaration order for performance reasons (otherwise, the same result may be cached multiple times
#       e.g. f(a=1, b=2) and f(b=2, a=1)
#   - the most common 'users' are class methods
# ===================================================================================================================================
ARGS_KWARGS = (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)


def cache(f):
    sig = inspect.signature(f)
    params = sig.parameters
    num_args = len(params)
    if num_args == 0:
        return _cache_no_args(f)

    if num_args == 1:
        for param in params.values():
            if param.kind not in ARGS_KWARGS and param.default is inspect.Parameter.empty:
                return _cache_single_arg(f)

    return _cache_with_args(f)


def _cache_no_args(f):
    the_value = [XNone]

    def getter():
        v = the_value[0]
        if v is XNone:
            the_value[0] = v = f()
        return v

    def clear():
        the_value[0] = XNone

    getter.__name__ = f.__name__
    getter.value = the_value
    getter.clear = clear
    return getter


def _cache_single_arg(f):
    the_cache = {}

    def getter(arg):
        value = the_cache.get(arg, the_cache)  # -- will throw if arg is not hashable!
        if value is the_cache:  # -- i.e., arg is seen for the first time
            value = f(arg)
            the_cache[arg] = value

        return value

    getter.__name__ = f.__name__
    getter.cache = the_cache
    getter.clear = lambda: the_cache.clear()
    return getter


def _cache_with_args(f):
    the_cache = {}

    def getter(*args, **kwargs):
        key = *args, *tuple(kwargs.items())
        value = the_cache.get(key, the_cache)  # -- will throw if key is not hashable!
        if value is the_cache:  # -- i.e., key is seen for the first time
            value = f(*args, **kwargs)
            the_cache[key] = value

        return value

    getter.__name__ = f.__name__
    getter.cache = the_cache
    getter.clear = lambda: the_cache.clear()
    return getter


def standard_key(args: tuple, kwargs: dict) -> tuple:
    sorted_kwargs = tuple((k, kwargs[k]) for k in sorted(kwargs))
    return *args, *sorted_kwargs


# ========= The singleton


def ___operator_new(cls, *args, **kwargs):
    key = standard_key((cls, *args), kwargs)
    ptr = cls.___instances.get(key)
    if not ptr:
        ptr = object.__new__(cls)
        cls.___ctor(ptr, *args, **kwargs)
        cls.___instances[key] = ptr
    return ptr


def ___operator_reset(cls):
    cls.___instances = {}


def singleton(cls):
    cls.___instances = {}
    cls.__new__ = ___operator_new
    cls.___ctor = cls.__init__
    cls.__init__ = lambda self, *args, **kwargs: None
    cls._reset_singleton = lambda: ___operator_reset(cls)
    return cls


# def hash_key( obj ) -> int:
#     try:
#         return hash( obj )
#     except Exception:
#         return hash(str(obj))



from core_10x.directory import Directory

ANIMALS = Directory.define(
    'Animals',
    [
        'Microorganisms',
        [
            'Single Cell',
            'Multi Cell',
        ],
        'Mollusks',
        'Fishes',
        [
            'Salt Water',
            [
                'Beluga'
            ],
            'Fresh Water',
        ],
        'Amphibia',
        'Reptiles',
        'Birds',
        'Mammals',
        [
            'Cats',
            'Dogs',
            'Bears',
            'Whales',
            [
                'Bluewhale',
                'Orca',
                'Spermwhale',
                'Beluga'
            ],
        ],
    ],
)

FISH = Directory.define(
    ( 'FWF', 'Fresh Water Fish' ),
    [
        ( 'PkF',    'Pike Family' ),
        [
            ( 'NPK',    'Northern Pike' ),
            ( 'MSK',    'Muskie' ),
            ( 'PKL',    'Pickerel' ),
        ],
        ( 'PeF', 'Perch Family' ),
        [
            ( 'PCH',    'Common Perch' ),
            ( 'YPH',    'Yellow Perch' ),
            ( 'WLY',    'Walleye' ),
            ( 'SGR',    'Sagger' ),
        ],
        ( 'CrF',    'Carp Family' ),
        [
            ( 'CRP',    'Common Carp' ),
            ( 'WCP',    'Wild Carp' ),
            ( 'IDE',    'Ide' ),
            ( 'BRM',    'Bream' ),
            ( 'RCH',    'Roach' ),
        ],
    ],
)

STANDARD_PIXMAPS = Directory.define(
    'Standard Pixmaps',
    [
        'Title Bar',
        [
            'TitleBarMaxButton',
            'TitleBarCloseButton',
            'TitleBarNormalButton',
            'TitleBarShadeButton',
            'TitleBarUnshadeButton',
            'TitleBarContextHelpButton',
        ],
        'Message Box',
        [
            'MessageBoxInformation',
            'MessageBoxWarning',
            'MessageBoxCritical',
            'MessageBoxQuestion',
        ],
        'Drive',
        [
            'DriveFDIcon',
            'DriveHDIcon',
            'DriveCDIcon',
            'DriveDVDIcon',
            'DriveNetIcon',
        ],
        'Dir',
        [
            'DirHomeIcon',
            'DirOpenIcon',
            'DirClosedIcon',
            'DirIcon',
            'DirLinkIcon',
            'DirLinkOpenIcon',
        ],
        'File',
        [
            'FileIcon',
            'FileLinkIcon',
            'FileDialogStart',
            'FileDialogEnd',
            'FileDialogToParent',
            'FileDialogNewFolder',
            'FileDialogDetailedView',
            'FileDialogInfoView',
            'FileDialogContentsView',
            'FileDialogListView',
            'FileDialogBack',
        ],
        'Dialog',
        [
            'DialogOkButton',
            'DialogCancelButton',
            'DialogHelpButton',
            'DialogOpenButton',
            'DialogSaveButton',
            'DialogCloseButton',
            'DialogApplyButton',
            'DialogResetButton',
            'DialogDiscardButton',
            'DialogYesButton',
            'DialogNoButton',
            'DialogYesToAllButton',
            'DialogNoToAllButton',
            'DialogSaveAllButton',
            'DialogAbortButton',
            'DialogRetryButton',
            'DialogIgnoreButton',
        ],
        'Arrow',
        [
            'ArrowUp',
            'ArrowDown',
            'ArrowLeft',
            'ArrowRight',
            'ArrowBack',
            'ArrowRight',
            'ArrowForward',
        ],
        'Media',
        [
            'MediaPlay',
            'MediaStop',
            'MediaPause',
            'MediaSkipForward',
            'MediaSkipBackward',
            'MediaSeekForward',
            'MediaSeekBackward',
            'MediaVolume',
            'MediaVolumeMuted',
        ],
        'Other',
        [
            'DesktopIcon',
            'TrashIcon',
            'ComputerIcon',
            'DockWidgetCloseButton',
            'ToolBarHorizontalExtensionButton',
            'ToolBarVerticalExtensionButton',
            'CommandLink',
            'VistaShield',
            'BrowserReload',
            'BrowserStop',
            'LineEditClearButton',
            'RestoreDefaultsButton',
        ],
    ],
    #'CustomBase',
)

if __name__ == '__main__':
    h1 = FISH.choices()

    f = ANIMALS.flatten()
    h2 = ANIMALS.choices()



from __future__ import annotations

import bisect
import functools
from typing import Callable
import inspect
from types import GeneratorType
import itertools

from core_10x.traitable import Traitable, Trait, T, RT, RC, RC_TRUE, XNone
from core_10x.named_constant import NamedConstant, NamedCallable
from core_10x.xinf import XInf

class Bucket(Traitable, root_class = True, embeddable = True):
    def calc_trait_values(self, trait_name: str, aggregator_f: Callable):
        data_gen = ( (member.get_value(trait_name), qty) for member, qty in self.members_qtys() )
        return data_gen if not aggregator_f else aggregator_f(data_gen)

    def calc_trait_values_with_args(self, trait_name: str, aggregator_f: Callable, *args):
        data_gen = ( (member.get_value(trait_name, *args), qty) for member, qty in self.members_qtys() )
        return data_gen if not aggregator_f else aggregator_f(data_gen)

    def calc_method(self, method_name: str, aggregator_f: Callable, *args, **kwargs):
        data_gen = ( (getattr(member, method_name)(*args, **kwargs), qty) for member, qty in self.members_qtys() )
        return data_gen if not aggregator_f else aggregator_f(data_gen)


    def _insert(self, obj: Traitable, qty: float):      raise NotImplementedError
    def _insert_bucket(self, bucket: Bucket):           raise NotImplementedError
    def is_member(self, obj: Traitable) -> bool:        raise NotImplementedError
    def members(self):                                  raise NotImplementedError
    def members_qtys(self):                             raise NotImplementedError

class BucketSet(Bucket):
    data: set[Traitable] = T(T.STICKY)

    def _insert(self, obj: Traitable, qty: float):
        self.data.add(obj)

    def _insert_bucket(self, bucket: BucketSet):
        self.data.update(bucket.data)

    def is_member(self, obj: Traitable) -> bool:
        return obj in self.data

    class Iter:
        def __init__(self, members: set):
            self.it = iter(members)

        def __iter__(self):
            return self

        def __next__(self):
            return (self.it.__next__(), 1.)

    def members(self):
        return iter(self.data)

    def members_qtys(self):
        return self.Iter(self.data)

class BucketList(Bucket):
    data: list = T(T.STICKY)

    def _insert(self, obj: Traitable, qty: float):
        self.data.append(obj)

    def _insert_bucket(self, bucket: BucketList):
        self.data.extend(bucket.data)

    def is_member(self, obj: Traitable) -> bool:
        return obj in self.data

    class Iter:
        def __init__(self, members: list):
            self.it = iter(members)

        def __iter__(self):
            return self

        def __next__(self):
            return (self.it.__next__(), 1.)

    def members(self):
        return iter(self.data)

    def members_qtys(self):
        return self.Iter(self.data)

class BucketDict(Bucket):
    data: dict[Traitable, float] = T(T.STICKY)

    def _insert(self, obj: Traitable, qty: float = 1):
        data = self.data
        ex_qty = data.get(obj, 0.)
        data[obj] = ex_qty + qty

    def _insert_bucket(self, bucket: BucketDict):
        data = self.data
        for obj, qty in bucket.data.items():
            ex_qty = data.get(obj, 0.)
            data[obj] = ex_qty + qty

    def is_member(self, obj: Traitable) -> bool:
        return obj in self.data

    def members(self):
        return self.data.keys()

    def members_qtys(self):
        return self.data.items()

class BUCKET_SHAPE(NamedConstant):
    SET     = BucketSet
    DICT    = BucketDict
    LIST    = BucketList

class Basketable:
    s_bucket_shape: BUCKET_SHAPE = None
    def __init_subclass__(cls, bucket_shape: BUCKET_SHAPE = None, **kwargs):
        if bucket_shape is not None:
            cls.s_bucket_shape = bucket_shape

        super().__init_subclass__(**kwargs)

    def contents(self, basket: Basket, qty: float = 1.):
        target_class = basket.base_class
        for member, mem_qty in self.members_qtys():
            if isinstance(member, target_class):
                basket.add(member, qty * mem_qty)
            else:
                member.contents(basket, qty = qty * mem_qty)


    def is_member(self, obj: Basketable) -> bool:   raise NotImplementedError
    def members_qtys(self):                         raise NotImplementedError


class Bucketizer(Traitable, embeddable = True):
    buckets_spec: list  = T()
    bucket_tags: set    = RT()

    def bucket_tags_get(self) -> set:
        return set(self.buckets_spec)

    def calc_bucketizing_value(self, obj: Traitable):
        raise NotImplementedError

    def calc_bucket_tag(self, bucketizing_value):
        known_tags = self.bucket_tags
        return bucketizing_value if not known_tags or bucketizing_value in known_tags else None

    @classmethod
    def verify_base_class(cls, base_class):
        if not base_class or not inspect.isclass(base_class) or not issubclass(base_class, Traitable):
            raise AssertionError('base_class must be a subclass of Traitable')

    @classmethod
    def verify_custom_f(cls, custom_f, label: str, arg_data_type: type = None):
        if isinstance(custom_f, NamedCallable):
            custom_f = custom_f.value

        if not callable(custom_f):
            raise AssertionError(f'{label}: {custom_f} is not callable')

        if arg_data_type is not None:
            sig = inspect.signature(custom_f)
            params = list(sig.parameters.values())
            if len(params) != 1:
                raise AssertionError(f'{label}: {custom_f} must accept exactly one parameter')
            p = params[0]
            dtype = p.annotation
            if not dtype is inspect.Signature.empty and not dtype is arg_data_type:
                raise AssertionError(f'{label}: {custom_f} must accept single {arg_data_type}')

    @classmethod
    def by_class(cls, base_class: type[Traitable], *known_subclasses) -> Bucketizer:
        cls.verify_base_class(base_class)
        if known_subclasses:
            if not all(inspect.isclass(ca) and issubclass(ca, base_class) for ca in known_subclasses):
                raise AssertionError(f'Each class in subclasses_allowed must be a subclass of {base_class}')

        return BucketizerByClass(base_class = base_class, buckets_spec = list(known_subclasses))

    @classmethod
    def by_feature(cls, base_class: type[Traitable], feature_calc, *buckets_spec, bucket_tag_calc = None) -> Bucketizer:
        cls.verify_base_class(base_class)
        cls.verify_custom_f(feature_calc, 'feature_calc')
        if not bucket_tag_calc:
            return BucketizerByFeature(base_class = base_class, f_bucketizing_value = feature_calc, buckets_spec = buckets_spec)

        cls.verify_custom_f(bucket_tag_calc, 'bucket_tag_calc')
        return BucketizerByFeature(base_class = base_class, f_bucketizing_value = feature_calc, f_bucket_tag = bucket_tag_calc, buckets_spec = buckets_spec)

    @classmethod
    def by_range(cls, base_class: type[Traitable], value_for_range_calc, *intervals_spec) -> Bucketizer:
        cls.verify_base_class(base_class)
        cls.verify_custom_f(value_for_range_calc, 'value_for_range_calc')
        return BucketizerByRange(base_class = base_class, f_bucketizing_value = value_for_range_calc, buckets_spec = intervals_spec)

    @classmethod
    def by_breakpoints(cls, base_class: type[Traitable], value_for_range_calc, *breakpoints) -> Bucketizer:
        cls.verify_base_class(base_class)
        cls.verify_custom_f(value_for_range_calc, 'value_for_range_calc')
        return BucketizerByBreakPoints(base_class = base_class, f_bucketizing_value = value_for_range_calc, buckets_spec = breakpoints)


class BucketizerByClass(Bucketizer):
    #def is_acceptable(self, obj: Traitable) -> bool:
    #    return isinstance(obj, self.base_class)

    def calc_bucketizing_value(self, obj: Traitable):
        return obj.__class__

class BucketizerByFeature(Bucketizer):
    f_bucketizing_value: NamedCallable  = T(T.NOT_EMPTY)
    f_bucket_tag: NamedCallable         = T()

    def calc_bucketizing_value(self, obj: Traitable):
        return self.f_bucketizing_value.value(obj)

    def calc_bucket_tag(self, bucketizing_value):
        f_bucket_tag = self.f_bucket_tag
        return f_bucket_tag.value(bucketizing_value) if f_bucket_tag else super().calc_bucket_tag(bucketizing_value)

class Interval:
    def __init__(self, a, b, inclusive = True, label: str = None):
        if inclusive:
            self.test_f = lambda v: a <= v and v <= b
        else:
            self.test_f = lambda v: a <= v and v < b

        if label is None:
            left_bracket = '(' if a is -XInf else '['
            right_bracket = ')' if b is XInf or not inclusive else ']'
            label = f'{left_bracket} {a} : {b} {right_bracket}'
        self.label = label

class BucketizerByRange(BucketizerByFeature):
    """
    interval    := [ label, low, high ]         #-- low <= x <= high: 'label'
                |= [ low, high ]                #-- same as [ '[low - high]', low, high ]
                |= ( label, low, high )         #-- low <= x < high: 'label'
                |= ( low, high )                #-- same as ( '[low - high )', low, high )
    """
    intervals: list[Interval]   = RT()

    def intervals_get(self) -> list:
        res = []
        for spec in self.buckets_spec:
            if isinstance(spec, list):
                inclusive = True
            elif isinstance(spec, tuple):
                inclusive = False
            else:
                raise TypeError('Each interval must either be list or tuple')

            n = len(spec)
            if n < 2 or n > 3:
                raise ValueError('Each interval must be either <label, low, high> or <low, high>')

            if n == 3:
                label, low, high = spec
                interval = Interval(low, high, inclusive = inclusive, label = label)
            else:
                low, high = spec
                interval = Interval(low, high, inclusive = inclusive)

            res.append(interval)

        return res

    def calc_bucket_tag(self, bucketizing_value) -> str:
        interval: Interval
        for interval in self.intervals:
            if interval.test_f(bucketizing_value):
                return interval.label

        return None

class BucketizerByBreakPoints(BucketizerByFeature):
    include_last: bool          = T(False)

    intervals: list[Interval]   = RT()

    def intervals_get(self) -> list:
        points = self.buckets_spec
        n = len(points)
        if n < 2:
            return []

        res = []
        include_last = self.include_last
        low = points[0]
        for i in range(1, n):
            point = points[i]
            if point is None:
                raise ValueError('None is not a valid point')

            try:
                rc = low < point
            except Exception:
                raise ValueError(f'Uncomparable points detected: ({low}, {point})')

            if not rc:
                raise ValueError(f'Points must be in strict ascending order: ({low}, {point})')

            inclusive = include_last if i == n - 1 else True
            interval = Interval(low, point, inclusive = inclusive)
            res.append(interval)

            low = point

        return res

    def calc_bucket_tag(self, bucketizing_value) -> str:
        break_points = self.buckets_spec
        intervals = self.intervals

        i = bisect.bisect_right(break_points, bucketizing_value) - 1
        if i < 0:
            return None

        if i == len(intervals):
            if not self.include_last:
                return None

            last = intervals[-1]
            return last.label if bucketizing_value == break_points[-1] else None

        return intervals[i].label

"""
Example of aggregator:

class FIN_AGGREGATOR(NamedCallable):
    PRICE       = FinInstrument.aggregate_price
    PRICE_CCY   = Portfolio.aggregate_price
    LIFE_CYCLE  = LifeCycler.aggregate_life_cycle
    LEAVES      = lambda basket:    raise NotSupportedError 
"""
class Basket(Traitable, embeddable = True):
    s_bucket_shape: BUCKET_SHAPE = BUCKET_SHAPE.DICT
    def __init_subclass__(cls, bucket_shape: BUCKET_SHAPE = None, **kwargs):
        super().__init_subclass__(**kwargs)
        if bucket_shape is not None:
            cls.s_bucket_shape = bucket_shape

    bucket_shape: BUCKET_SHAPE              = RT()

    base_class: type[Traitable]             = T(T.NOT_EMPTY)
    subclasses_allowed: bool                = T(True)
    aggregator_class: type[NamedCallable]   = T()
    bucketizers: list[Bucketizer]           = T(T.EMBEDDED)

    the_bucket: Bucket                      = T(T.STICKY)       #-- single bucket if there are no bucketizers
    all_buckets: dict                       = T(T.STICKY)       #-- tagged buckets WRT bucketizers, i.e.: {(t1_i,t2_i,...): bucket_i}


    def bucket_shape_get(self) -> BUCKET_SHAPE:
        base_class = self.base_class
        if issubclass(base_class, Basketable):
            return base_class.s_bucket_shape
        return self.__class__.s_bucket_shape

    def the_bucket_get(self) -> Bucket:
        return self.new_bucket()

    def bucketizers_set(self, trait, bucketizers: list[Bucketizer]) -> RC:
        if not isinstance(bucketizers, list) or not all(isinstance(b, Bucketizer) for b in bucketizers):
            return RC(False, 'bucketizers must be a list of instances of Bucketizer')

        self.invalidate_value('the_bucket')
        self.invalidate_value('all_buckets')

        return self.raw_set_value(trait, bucketizers)

    def add_bucketizer(self, bucketizer: Bucketizer) -> bool:
        new_buckets = {}
        for tag, bucket in self.tags_buckets():
            for member, qty in bucket.members_qtys():
                b_value = bucketizer.calc_bucketizing_value(member)
                b_tag = bucketizer.calc_bucket_tag(b_value)
                if b_tag is None:
                    continue

                key = *tag, b_tag
                new_bucket = new_buckets.get(key)
                if new_bucket is None:
                    new_bucket = self.new_bucket()
                    new_buckets[key] = new_bucket
                new_bucket._insert(member, qty)

        self.invalidate_value('the_bucket')
        self.raw_set_value('all_buckets', new_buckets)
        self.bucketizers.append(bucketizer)
        return True

    def new_bucket(self) -> Bucket:
        return self.bucket_shape.value()

    def is_acceptable(self, obj: Traitable) -> bool:
        return isinstance(obj, self.base_class) if self.subclasses_allowed else obj.__class__ is self.base_class

    def add(self, obj: Traitable, qty: float = 1.) -> bool:
        if not self.is_acceptable(obj):
            return False

        bucketizers = self.bucketizers
        if not bucketizers:
            bucket = self.the_bucket
        else:
            tags = []
            for bucketizer in bucketizers:
                b_value = bucketizer.calc_bucketizing_value(obj)
                b_tag = bucketizer.calc_bucket_tag(b_value)
                if b_tag is None:
                    return False

                tags.append(b_tag)

            key = tuple(tags)
            data = self.all_buckets
            bucket = data.get(key)
            if bucket is None:
                bucket = self.new_bucket()
                data[key] = bucket

        bucket._insert(obj, qty)
        return True

    def tags_buckets(self):
        if not self.bucketizers:
            return ( v for v in ((XNone, self.the_bucket),) )

        return ( (key, bucket) for key, bucket in self.all_buckets.items() )

    def buckets_by_subtags(self, *subtags):
        n_bucks = len(self.bucketizers)
        n_subtags = len(subtags)
        if n_subtags > n_bucks:
            raise AssertionError(f"Can't specify more than {n_bucks} subtags")

        if n_bucks == 0:
            return self.tags_buckets()

        look_for = n_bucks * [None]
        for i, s in enumerate(subtags):
            look_for[i] = s

        for tag, bucket in self.tags_buckets():
            if all(tag[i] == sub for i, sub in enumerate(look_for) if sub is not None):
                yield (tag, bucket)

    def members_qtys(self):
        return itertools.chain.from_iterable(bucket.members_qtys() for _, bucket in self.tags_buckets())

    def __getattr__(self, method_name: str):
        base_class = self.base_class
        trait = getattr(base_class, method_name, None)
        if trait is None:
            return None

        aggregator_f = self.f_aggregator(method_name, throw = False)
        if isinstance(trait, Trait):
            if not trait.getter_params:
                return self.calc_trait_values(method_name, aggregator_f)

            return functools.partial(self._lift_by_trait_name, method_name, aggregator_f)

        if callable(trait):
            return functools.partial(self._lift_by_method_name, method_name, aggregator_f)

        return None

    def calc_trait_values(self, trait_name: str, aggregator_f: Callable):
        if not self.bucketizers:
            return self.the_bucket.calc_trait_values(trait_name, aggregator_f)

        return {
            tag: bucket.calc_trait_values(trait_name, aggregator_f)
            for tag, bucket in self.tags_buckets()
        }

    def _lift_by_trait_name(self, trait_name: str, aggregator_f: Callable, *args):
        if not self.bucketizers:
            return self.the_bucket.calc_trait_values_with_args(trait_name, aggregator_f, *args)

        return {
            tag: bucket.calc_trait_values_with_args(trait_name, aggregator_f, *args)
            for tag, bucket in self.tags_buckets()
        }

    def _lift_by_method_name(self, method_name: str, aggregator_f: Callable, *args, **kwargs):
        if not self.bucketizers:
            return self.the_bucket.calc_method(method_name, aggregator_f, *args, **kwargs)

        return {
            tag: bucket.calc_method(method_name, aggregator_f, *args, **kwargs)
            for tag, bucket in self.tags_buckets()
        }

    def finalize_results(self, results):
        if isinstance(results, GeneratorType):
            return list(results)

        if isinstance(results, dict):
            return { tag: list(r) if isinstance(r, GeneratorType) else r for tag, r in results.items() }

        return results

    def f_aggregator(self, method_name: str, throw = True) -> Callable:
        f = self.aggregator_class.s_dir.get(method_name.upper())
        if throw and f is None:
            raise AssertionError(f"Basket: aggregator for method '{method_name}' is not defined")
        return f.value




from datetime import date

from infra_10x.mongodb_store import MongoStore

from core_10x.code_samples.person import Person
from core_10x.exec_control import GRAPH_ON

if __name__ == '__main__':

    with MongoStore.instance(hostname ='localhost', dbname='test', username='', password=''):
        p = Person(first_name = 'Ilya', last_name = 'Pevzner')
        p.dob = date(1971,7,1)
        p.save()
        p.dob = date(1971,7,3)
        assert p.dob== date(1971,7,3)
        with GRAPH_ON():
            assert p.dob == date(1971,7,3)  # FIX: fails here!



import pytest

from core_10x.ts_store import TsStore


@pytest.fixture(scope='module')
def ts_instance():
    from core_10x.testlib.test_store import TestStore

    assert not TsStore.s_instances
    yield TestStore.instance()
    TsStore.s_instances.clear()



from __future__ import annotations

from typing import Any, Callable

from core_10x.nucleus import Nucleus


#-- TODO: should I re-incarnate deriving from a metaclass, so that I can do: NAMED_CONSTANT[name] ?
class NamedConstant(Nucleus):
    """
    In a subclass, each line representing a member must be one of the following:
        MEMBER_NAME = (label, value)    #-- label,          value
        MEMBER_NAME = (label, )         #-- label,          auto-value
        MEMBER_NAME = ()                #-- MEMBER_NAME,    auto-value
        MEMBER_NAME = value             #-- MEMBER_NAME,    value

        where:

        MEMBER_NAME - a valid uppercase identifier
        label       - a string
        value       - any value
        auto-value  - automatically generated (int) value

        NOTE:   all the members must have either explicit or auto-generated values exclusively
    """

    __slots__ = 'label', 'name', 'value'

    def __init__(self, name: str = '', label: str = '', value: Any = None):
        super().__init__()
        self.name = name
        self.label = label
        self.value = value

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def __deepcopy__(self, memo=None):
        return self

    # -- Makes sense only for NamedConstants with Callable values
    def __call__(self, *args, **kwargs):
        return self.value(*args, **kwargs)

    def is_member_by_name(self, other_class) -> bool:
        return self.name in other_class.s_dir

    @classmethod
    def _create(cls, args: Any) -> NamedConstant:
        cdef = cls()
        if type(args) is not tuple:  #-- just a value
            cdef.value = args

        else:
            n = len(args)
            if n == 0:  #-- ()
                pass

            elif n == 1:  #-- just a label
                cdef.label = args[0]

            elif n == 2:  #-- (label, value)
                cdef.label = args[0]
                cdef.value = args[1]

            else:
                raise ValueError(f"'{args}' may have at most two items")

        return cdef

    #===================================================================================================================
    #   Nucleus Interface implementation
    #===================================================================================================================

    def to_str(self) -> str:
        return f'{self.__class__.__name__}.{self.name}'

    def to_id(self) -> str:
        return self.name

    def serialize(self, embed: bool) -> str:
        return self.name

    @classmethod
    def deserialize(cls, name: str) -> NamedConstant:
        sdir = cls.s_dir
        cdef = sdir.get(name, sdir)
        if cdef is sdir:
            raise TypeError(f'{cls} - unknown constant {name}')

        return cdef

    @classmethod
    def from_str(cls, s: str) -> NamedConstant:
        return cls.s_dir.get(s)

    @classmethod
    def from_any_xstr(cls, data) -> NamedConstant | None:
        if type(data) is cls.s_data_type:
            reverse_dir = cls.s_reverse_dir
            if reverse_dir:
                return reverse_dir.get(data)

            #-- the last resort - this is slow!
            for cdef in cls.s_dir.values():
                if cdef.value == data:
                    return cdef
        return None

    @classmethod
    def same_values(cls, value1, value2) -> bool:
        return value1 is value2

    @classmethod
    def choose_from(cls):
        return cls.s_dir

    #===================================================================================================================

    s_dir: dict[str, NamedConstant] = {}
    s_reverse_dir = {}
    s_data_type = None
    s_default_labels = False
    s_lowercase_values = False

    def __init_subclass__(
        cls,
        default_labels: bool    = None,    #-- if True and a label is not defined, creates it by calling default_label(name)
        lowercase_values: bool  = None,    #-- if a value is not defined, sets it to name, or name.lower() if True
        data_type: type         = None,    #-- if value is not defined it is taken from superclass or the first value
    ):
        sdir = cls.s_dir = {name: cls(cdef.name, cdef.label, cdef.value) for name, cdef in cls.s_dir.items()}

        if default_labels is not None:
            cls.s_default_labels = default_labels
        else:
            default_labels = cls.s_default_labels

        if lowercase_values is not None:
            cls.s_lowercase_values = lowercase_values
        else:
            lowercase_values = cls.s_lowercase_values

        if data_type is not None:
            cls.s_data_type = data_type
        else:
            data_type = cls.s_data_type

        for name, constant_definition in cls.__dict__.items():
            if not name.isupper():
                continue

            cdef = cls._create(constant_definition)
            cdef.name = name
            if not cdef.label:
                cdef.label = cls.default_label(name) if default_labels else name

            if cdef.value is None:
                value = cls.next_auto_value()
                if value is None:
                    value = name if not lowercase_values else name.lower()
                cdef.value = value

            dt = type(cdef.value)
            if data_type is None:
                data_type = dt
            else:
                assert issubclass(dt, data_type), f'{cls}.{name} must be a subclass of {data_type}'

            sdir[name] = cdef

        cls.s_data_type = data_type
        if getattr(data_type, '__hash__', None):    #-- check if data_type is hashable and build the reverse dir if so
            cls.s_reverse_dir = {cdef.value: cdef for cdef in sdir.values()}

        for name, cdef in sdir.items():
            setattr(cls, name, cdef)

    @staticmethod
    def union(*named_constant_classes):
        class _union(NamedConstant):
            pass
        dir = _union.s_dir
        for cls in named_constant_classes:
            dir.update(cls.s_dir)

        for name, value in dir.items():
            setattr(_union, name, value)

        return _union

    @classmethod
    def default_label(cls, name: str) -> str:
        return name.title().replace('_', ' ')

    @classmethod
    def next_auto_value(cls):
        return None

    @classmethod
    def item(cls, symbol_name: str) -> NamedConstant:
        return cls.s_dir.get(symbol_name)

class NamedCallable(NamedConstant):
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        for cvar in cls.s_dir.values():
            assert callable(cvar.value), f'{cls}.{cvar.name} - value must be Callable'
            break   #-- the types of all the rest must have been checked in super()

    @staticmethod
    def just_func(f: Callable):
        return NamedCallable(None, None, f)

class Enum(NamedConstant):
    """
    Enum constants definitions are in a simplified form:
        - an empty tuple (next enum auto value)
        - a string - just a label
    """

    s_last_value = 0
    s_step = 1

    def __init_subclass__(cls, seed: int = None, step: int = None, **kwargs):
        if seed is not None:
            cls.s_last_value = seed
        else:
            cls.s_last_value = cls.s_last_value

        if step is not None:
            cls.s_step = step
        else:
            cls.s_step = cls.s_step

        super().__init_subclass__(**kwargs)

    def __int__(self):
        return self.value

    @classmethod
    def _create(cls, args: Any) -> Enum:
        cdef = cls()
        if args == ():
            return cdef

        if type(args) is str:  #-- just a label
            cdef.label = args
            return cdef

        raise ValueError('an empty tuple or string is expected')

    @classmethod
    def next_auto_value(cls) -> int:
        last_value = cls.s_last_value
        cls.s_last_value += cls.s_step
        return last_value


NO_FLAGS_TAG = 'NONE'

class EnumBits(NamedConstant):
    s_last_flag = 0x1

    def __init_subclass__(cls, **kwargs):
        cls.s_last_flag = cls.s_last_flag
        super().__init_subclass__(**kwargs)
        setattr(cls, NO_FLAGS_TAG, cls(name = NO_FLAGS_TAG, value = 0x0))

    @classmethod
    def next_auto_value(cls):
        mask = cls.s_last_flag
        cls.s_last_flag <<= 1
        return mask

    @classmethod
    def names_from_value(cls, value: int) -> tuple:
        return tuple(name for name, cdef in cls.s_dir.items() if cdef.value & value)

    def __or__(self, other):
        the_value = self.value
        value = the_value | other.value
        if value == the_value:
            return self

        cls = self.__class__
        return cls.from_int(value)

    def __and__(self, other):
        the_value = self.value
        value = the_value & other.value
        if value == the_value:
            return self

        cls = self.__class__
        return cls.from_int(value)

    def __sub__(self, other):
        the_value = self.value
        value = the_value & ~other.value
        if value == the_value:
            return self

        cls = self.__class__
        return cls.from_int(value)

    @classmethod
    def names_to_value(cls, cnames: list) -> int:
        value = 0x0
        sdir = cls.s_dir
        for cname in cnames:
            cdef = sdir.get(cname)
            if cdef is None:
                raise TypeError(f'{cls} - unknown bit {cname}')

            value |= cdef.value

        return value

    @classmethod
    def from_str(cls, s: str) -> EnumBits:
        if not s:
            return getattr(cls, NO_FLAGS_TAG)

        found = cls.s_dir.get(s)
        if found is not None:
            return found

        cnames = s.split('|')
        value = cls.names_to_value(cnames)
        return cls(s, s, value)

    @classmethod
    def from_int(cls, value: int) -> EnumBits:
        cnames = cls.names_from_value(value)
        if not cnames:
            return getattr(cls, NO_FLAGS_TAG)

        name = '|'.join(cnames)
        return cls(name, name, value)

    @classmethod
    def from_any_xstr(cls, data) -> EnumBits:
        dt = type(data)
        if dt is int:
            return cls.from_int(data)

        if dt is tuple or dt is list:
            value = cls.names_to_value(data)
            if not value:
                return getattr(cls, NO_FLAGS_TAG)

            name = '|'.join(data)
            return cls(name, name, value)
        return None

    @classmethod
    def deserialize(cls, data) -> EnumBits:
        return cls.from_str(data)

    @classmethod
    def same_values(cls, value1, value2) -> bool:
        return value1 is value2 or value1.name == value2.name


class ErrorCode(Enum, seed=-1, step=-1):
    def __call__(self, *args, **kwargs):
        return self.label.format(*args, **kwargs)


class NamedConstantValue:
    __slots__ = 'data', 'named_constant_class'

    def __init__(self, named_constant_class, _check_num_values = True, **named_constant_values):
        assert issubclass(named_constant_class, NamedConstant), f'{named_constant_class} must be a subclass of NamedConstant'
        c_defs = named_constant_class.s_dir

        if _check_num_values:
            num_values = len(c_defs)
            assert num_values == len(named_constant_values), f'{named_constant_class} - number of named values must be {num_values}'

        self.named_constant_class = named_constant_class

        self.data = data = {}
        for cname, value in named_constant_values.items():
            cdef = c_defs.get(cname)
            assert cdef, f'{named_constant_class}.{cname} - unknown named constant'
            self.process_row(cdef, data, value)

    def process_row(self, cdef: NamedConstant, data: dict, row):
        data[cdef] = row

    def __getitem__(self, key):
        try:
            return self.data[key]  #-- if it's a known NamedConstant

        except Exception as e:
            #-- check if it is a name of a constant
            named_constant_class = self.named_constant_class
            cdef = named_constant_class.s_dir.get(key)
            if not cdef:
                raise KeyError(f'{named_constant_class}.{key} - unknown named constant') from e

            return self.data[cdef]

    def __setitem__(self, key, value):
        raise AssertionError(f'{self.__class__} - may not be modified')

    def __getattr__(self, key):
        return self.__getitem__(key)


class NamedConstantTable(NamedConstantValue):
    __slots__ = 'col_named_constant_class', 'data', 'named_constant_class'

    def __init__(self, row_nc_class, col_nc_class, _check_num_values = True, **named_tuple_values):
        assert issubclass(col_nc_class, NamedConstant), f'{col_nc_class} must be a subclass of NamedConstant'
        self.col_named_constant_class = col_nc_class
        super().__init__(row_nc_class, _check_num_values = _check_num_values, **named_tuple_values)

    def process_row(self, cdef: NamedConstant, data: dict, row):
        assert type(row) is tuple, f'{cdef.name} = is not a tuple'
        col_defs = self.col_named_constant_class.s_dir
        assert len(row) == len(col_defs), f'{cdef.name} must have {len(col_defs)} values'
        data[cdef] = NamedConstantValue(self.col_named_constant_class, **{col_name: row[i] for i, col_name in enumerate(col_defs)})

    def primary_key(self, sec_key, value):
        for primary_key, row in self.data.items():
            if row[sec_key] == value:
                return primary_key

        return None

    def extend(self, row_nc_subclass: type[NamedConstant], **named_tuple_values) -> NamedConstantTable:
        assert issubclass(row_nc_subclass, self.named_constant_class), f'row_nc_subclass must be a subclass of {self.named_constant_class}'

        res = NamedConstantTable(row_nc_subclass, self.col_named_constant_class, _check_num_values = False)
        res.data = data = dict(self.data)

        num_values = len(row_nc_subclass.s_dir) - len(self.named_constant_class.s_dir)
        assert len(named_tuple_values) == num_values, f'{row_nc_subclass} - number of named values must be {num_values}'

        c_defs = row_nc_subclass.s_dir
        for cname, value in named_tuple_values.items():
            cdef = c_defs.get(cname)
            assert cdef, f'{row_nc_subclass}.{cname} - unknown named constant'
            res.process_row(cdef, data, value)

        return res



from core_10x.directory import Directory
from core_10x.named_constant import Enum
from core_10x.xnone import XNone

from ui_10x.utils import UxDialog, UxRadioBox, UxSearchableList, UxTreeViewer, ux


class Choice:
    """
    choices     - a dict of labels to tags values or Directory; if given, f_choices is ignored
    f_choices   - a fully bound function that must return choices (see above)

    Q1: do we want to allow setting tags_selected with tags outside of all_choices? If so, what about updating all_choices
        and potentially saving such Choice instance?
    """
    def __init__(self, choices = XNone, f_choices = None, f_selection_callback = None, **kwargs):
        self.choices: dict = XNone
        self.inverted_choices: dict = XNone
        self.f_choices = None
        self.f_selection_cb = None
        self.directory: Directory = None
        self.kwargs = {}

        self.values_selected = []

        self.set(choices = choices, f_choices = f_choices, f_selection_callback = f_selection_callback, **kwargs)

    def set(self, choices = XNone, f_choices = None, f_selection_callback = None, **kwargs):
        assert choices or f_choices, 'Either choices or f_choices must be given'

        self.f_selection_cb = f_selection_callback

        if (choices):
            self.f_choices = None
        else:
            self.f_choices = f_choices
            choices = f_choices()       #-- TODO: we may want to wrap it in try - except

        self.kwargs.update(kwargs)

        if isinstance(choices, Directory):
            self.directory = choices
            self.choices = choices.choices()

        elif isinstance(choices, dict):
            self.directory = None
            self.choices = choices

        else:   #-- must be an iterable then!
            self.directory = None
            try:
                self.choices = {hint: hint for hint in choices}
            except Exception:
                self.choices = {}

        self.inverted_choices = {value: label for label, value in self.choices.items()}

    def widget(self) -> ux.Widget:
        if self.directory:
            return UxTreeViewer(
                self.directory,
                select_hook = lambda dir: self.on_item_selected(dir, dir.value, False),
                **self.kwargs
            )

        return UxSearchableList(
            choices     = list(self.choices.keys()),
            select_hook = lambda item_value: self.on_item_selected(None, item_value, True),
            **self.kwargs
        )

    def on_item_selected(self, dir: Directory, item, convert_choice: bool):
        item_value = self.choices.get(item) if convert_choice else item
        if item_value is not None:
            self.on_value_selected(dir, item_value, convert_value = convert_choice)
            if self.f_selection_cb:
                self.f_selection_cb(item)

    def on_value_selected(self, dir: Directory, item_value, convert_value: bool = False):
        self.values_selected = [item_value]

    def _on_selection_in_dialog(self, d: UxDialog, current_selection_cb, item_value):
        if current_selection_cb:
            current_selection_cb(item_value)
        d.done(1)
        self.f_selection_cb = current_selection_cb

    def popup(self, parent: ux.Widget = None, _open = True) -> UxDialog:
        _d = UxDialog(
            self.widget(),
            parent = parent,
            title = 'Pick your choice',
            ok = '', cancel = '',
            #window_flags = Qt.FramelessWindowHint | Qt.Window | Qt.CustomizeWindowHint
        )

        current_selection_cb = self.f_selection_cb
        self.f_selection_cb = lambda item_value: self._on_selection_in_dialog(_d, current_selection_cb, item_value)

        if _open:
            _d.show()

        return _d


class MultiChoice(Choice):
    class SELECT_MODE(Enum):
        ANY     = ()
        LEAF    = ()
        SUBDIR  = ()

    def __init__(self, *choices_selected, choices = XNone, f_choices = None, **kwargs):
        super().__init__(choices = choices, f_choices = f_choices, **kwargs)
        self._set_choices_selected(choices_selected)

    def set(self, *choices_selected, choices = XNone, f_choices = None, **kwargs):
        super().set(choices = choices, f_choices = f_choices, **kwargs)
        self._set_choices_selected(choices_selected)

    def _set_choices_selected(self, choices_selected: tuple):
        all_choices = self.choices
        self.values_selected.extend(tuple(value for choice in choices_selected if (value := all_choices.get(choice) )))

    def select_mode(self):
        return self.rb.choice() if self.rb else self.SELECT_MODE.ANY

    def widget(self) -> ux.Widget:
        self.sw = sw = super().widget()
        if not sw:
            return None

        w = ux.Splitter(ux.Horizontal)

        if isinstance(sw, UxTreeViewer):
            left = ux.Widget()
            left_lay = ux.VBoxLayout()
            left.set_layout(left_lay)

            self.rb = rb = UxRadioBox(self.SELECT_MODE, horizontal = True, title = 'Selection Mode')
            left_lay.add_widget(rb)
            left_lay.add_widget(sw)
        else:
            left = sw
            self.rb = None

        w.add_widget(left)

        self.selection_list = selection_list = ux.ListWidget()
        labels = self.values_selected
        if not self.directory:
            map = self.inverted_choices
            labels = [map.get(v) for v in labels if map.get(v)]

        selection_list.add_items(labels)
        selection_list.clicked_connect(self.on_selection_list_selection)
        w.add_widget(selection_list)

        return w

    def on_selection_list_selection(self, item: ux.ListItem):
        label = item.text()
        if not self.directory:
            value = self.choices.get(label)
        else:
            value = label

        assert value, f"Value for '{label}' must exist"

        self.values_selected.remove(value)
        row = self.selection_list.row(item)
        self.selection_list.take_item(row)

    def on_value_selected(self, subdir: Directory, item_value, convert_value = False):
        values_selected = self.values_selected
        if item_value not in values_selected:
            mode = self.select_mode()
            handler = self.s_select_table.get(mode)
            if handler(self, subdir, values_selected, item_value):
                values_selected.append(item_value)
                if convert_value:
                    item_value = self.inverted_choices.get( item_value )
                self.selection_list.add_item(item_value)

    def _select_leaf(self, subdir: Directory, values_selected: list, item_value) -> bool:
        return subdir.is_leaf()

    def _select_subdir(self, subdir: Directory, values_selected: list, item_value) -> bool:
        dir = self.directory
        assert dir, 'This is not a Directory'

        n = len(values_selected)
        for value in list(values_selected):
            if dir.is_value_contained(item_value, value):
                return False    #-- it's a subdir of something already selected

            if subdir.contains(value):
                values_selected.remove(value)

        if n > len(values_selected):    #-- one or more subdirs contained by the subdir got removed
            self.selection_list.clear()
            self.selection_list.add_items(values_selected)

        return True

    # fmt: off
    s_select_table = {
        SELECT_MODE.ANY:       lambda self, x,y,z: True,
        SELECT_MODE.LEAF:      _select_leaf,
        SELECT_MODE.SUBDIR:    _select_subdir
    }
    # fmt: on




from __future__ import annotations

from datetime import date

import rio
import ui_10x.platform_interface as i
from ui_10x.rio.component_builder import Widget


class CalendarWidget(Widget, i.CalendarWidget):
    s_component_class = rio.Calendar
    s_default_kwargs = dict(value=date.today())
    s_children_attr = 'value'
    s_single_child = True

    def set_grid_visible(self, grid_visible: bool):
        pass

    def set_selected_date(self, selected_date: date):
        self['value'] = selected_date

    def selected_date(self) -> date:
        return self['value']



from core_10x.named_constant import EnumBits


class STATE(EnumBits):
    RESERVED    = ()
    RUNTIME     = ()
    ID          = ()

class XSTATE(STATE):
    EXPENSIVE   = ()

if __name__ == '__main__':
    r = STATE.RUNTIME | STATE.RESERVED | STATE.ID

    r2 = r - STATE.RESERVED

    e = XSTATE.EXPENSIVE